#
#  crnverifier/pathway_decomposition.py
#  Original source from the Nuskell compiler project
#
#  Authors:
#   - Seung Woo Shin (seungwoo.theory@gmail.com)
#   - Stefan Badelt (bad-ants-fleet@posteo.eu)
#
import logging
log = logging.getLogger(__name__)

from itertools import chain
from collections import Counter
from .utils import assign_crn_species, pretty_crn, pretty_rxn, natural_sort, clean_crn, interpret

class NoFormalBasisError(Exception):
    pass

class BasisFinderError(Exception):
    pass

def formal(s, fs):
    """ Returns all formal species in state s. """
    assert isinstance(s, (list, chain, tuple))
    return [x for x in s if x in fs]

def intermediate(s, fs):
    """ Returns all non-formal species in state s. """
    assert isinstance(s, (list, chain, tuple))
    return [x for x in s if x not in fs]

def is_formal_state(s, fs):
    """ True if the state s is a formal state, False otherwise. """
    return len(intermediate(s, fs)) == 0

def is_subset(a, b):
    """ True if *a* is a subset of *b*, False otherwise. """
    assert isinstance(a, Counter) and isinstance(b, Counter)
    return a & b == a

class Path:
    """ A sequence of reactions.

    Can be initialized using a list of reactions in *list* or *Counter* format.
    Use __add__() to save computation time when concatinating two paths.
    """
    #TODO: Think about how we can save computation by expressing path decomposion
    #as a combination of path objects.
    def __init__(self, path, fs = None):
        assert isinstance(path, list)
        if len(path) == 0:
            self.lpath = self.cpath = []
        else:
            assert len(path[0]) == 2
            if isinstance(path[0][0], Counter):
                self.cpath = path
                self.lpath = [[list(R.elements()), list(P.elements())] for [R, P] in path]
            else: 
                self.cpath = [[Counter(R), Counter(P)] for [R, P] in path]
                self.lpath = path
        self.fs = fs # Formal species

        # set by: self._get_minits()
        self._minit = None
        self._final = None
        # set by: self.width
        self._width = None
        # set by: self.formal_closure
        self._formc = None
        # set by: self.states()
        self._cstates = None
        # set by: decomposed_final_states
        self._dfinals = None
        # set by: is_linear()
        self._linear = None
        # TODO
        self._dfs = None
        self._rfs = None

    @property
    def pretty_path(self):
        return map(pretty_rxn, self.lpath)

    @property
    def S0(self):
        """ tuple: hashable initial state. """
        return tuple(sorted(self.minimal_initial_state.elements()))

    @property
    def minimal_initial_state(self):
        """ Counter: minimal multiset of species to complete this path. """
        if self._minit is None:
            self._get_minits()
        return self._minit

    @property
    def Sn(self):
        """ tuple: hashable final state. """
        return tuple(sorted(self.final_state.elements()))

    @property
    def final_state(self):
        """ Counter: minimal multiset of species after completing this path. """
        if self._final is None:
            self._get_minits()
        return self._final

    @property
    def is_semiformal(self):
        """ bool: True if path has a formal initial state, False otherwise. """
        return is_formal_state(self.minimal_initial_state.elements(), self.fs)

    @property
    def is_formal(self):
        """ bool: True if path has formal initial and formal final state, False otherwise. """
        return is_formal_state(self.minimal_initial_state.elements(), self.fs) \
                and is_formal_state(self.final_state.elements(), self.fs)

    @property
    def is_prime(self):
        """ bool: True if path is undecomposable, False otherwise. """
        return len(self.dfs) == 0

    def is_linear(self, nonwastes):
        """ True if path has monomolecular substructure, False otherwise. 
        Args: nonwastes (list): non-formal reactants. """
        if self._linear is None:
            for state in self.states(counter = False):
                c1 = [x for x in state if x in nonwastes]
                if len(c1) > 1:
                    self._linear = False
                    break
            else:
                self._linear = True
        return self._linear

    @property
    def width(self):
        """ int: the largest state when folloing the path from minimal initial state. """
        if self._width is None:
            if len(self) == 0:
                self._width = 0
            else:
                self._width = max([sum(s.values()) for s in self.states(counter = True)]) 
        return self._width

    @property
    def fc(self):
        """ tuple: hashable formal closure. """
        return tuple(sorted(self.formal_closure.elements()))

    @property
    def formal_closure(self):
        """ Counter: minimal multiset of formal species involved in a path. """
        if self._formc is None:
            self._formc = Counter()
            for Si in self.states(counter = True):
                self._formc |= self.formal(Si)
        return self._formc

    @property
    def dfs(self):
        """ frozenset: a wrapper for decomposed final states. """
        return self.decomposed_final_states()

    @property
    def rfs(self):
        """ frozenset: a wrapper for regular final states. """
        return self.regular_final_states() # if len(self.dfs) == 0 else None
 
    def decomposed_final_states(self, path = None):
        """ Calculate decomposed final states.

        Note: If self._dfinal is None, then this routine always uses the
        internal path, otherwise it will update the internal final states with
        the given path.  This is a speedup used by the __add__ function.
        """
        if self.fs is None:
            raise BasisFinderError('DFS: no formal species given.')

        if self._dfinals is None:
            self._dfinals = [(Counter(), Counter())]
            assert path is None
            path = self.cpath
        elif path is None:
            path = []

        def next_state_C(S, rxn):
            """ Counter: applies reaction rxn to state S. """
            if not is_subset(rxn[0], S):
                raise BasisFinderError('Reaction {} cannot occur in state S {}.'.format(S, rxn))
            return S - rxn[0] + rxn[1]

        def next_final_states(t1, t2, R, P):
            try:
                tmp1 = t1 - self.formal(R) # consume formal species, if they are present.
                tmp1 = next_state_C(tmp1, [self.intermediate(R), P]) # skip formal reactants
            except BasisFinderError as err:
                tmp1 = None
            if t1 == t2:
                return [(tmp1, t2), None] if tmp1 is not None else [None, None]
            try:
                tmp2 = t2 - self.formal(R) # consume formal species, if they are present.
                tmp2 = next_state_C(tmp2, [self.intermediate(R), P]) # skip formal reactants
            except BasisFinderError as err:
                tmp2 = None
            if tmp1 is None and tmp2 is None:
                return None, None
            elif tmp2 is None:
                return (tmp1, t2), None
            elif tmp1 is None:
                return (t1, tmp2), None
            else:
                return (tmp1, t2), (t1, tmp2)

        debug = False
        new_finals = []
        for [R, P] in path:
            new_finals = []
            for (T1, T2) in self._dfinals:
                if debug: 
                     if (is_formal_state(T1.elements(), self.fs) and sum(T1.values()) != 0) or \
                        (is_formal_state(T2.elements(), self.fs) and sum(T2.values()) != 0):
                        raise BasisFinderError('DFS: found formal final state during decomposition!')
                n1, n2 = next_final_states(T1, T2, R, P)
                if n1 is not None and n1 not in new_finals[1:]:
                    new_finals.append((n1))
                if n2 is not None and n2 not in new_finals[1:]:
                    new_finals.append((n2))
            if len(new_finals) < 1:
                raise BasisFinderError('DFS: not a semiformal pathway!')
            self._dfinals = new_finals

        if len(self._dfinals) == 1:
            [(final, _)] = self._dfinals
            assert sum(_.values()) == 0 and final == self.final_state
            return frozenset()
        else:
            DFS = set()
            for e, (T1, T2) in enumerate(self._dfinals):
                if e == 0:
                    assert sum(T2.values()) == 0
                    (final, _) = (T1, T2)
                    assert final == self.final_state
                else:
                    DFS.add(tuple(sorted(T1.elements())))
                    DFS.add(tuple(sorted(T2.elements())))
            return frozenset(DFS)

    def regular_final_states(self):
        """ Calculates the set of final states assuming that the path is regular.

        A regular path must have a *turning point* reaction.

        Say a reaction (R,P)_t transforms state S_{t-1} to S_t. Fi is short for
        formal(Si).  A *turning point reaction* must fulfill several conditions:
              - every Fi with i<t is contained in S0 (= F0)
              - every Fi with i>=t is contained in final state Fn
              - formal(S_{t-1} - R_t) == formal(S_t - P_t) == 0

        A note for the last condition: "If there are formal species at state
        t-1, then those formal species have to be consumed by the turning
        point reaction". Equivalently, "if there are formal species at state t,
        then those species must have been produced by the turning point
        reaction".
       
        As example, a reaction (A -> A + i) does not change the formal species
        between S_{t-1} and S_t, but may be considered turning point reaction
        (if and only if A is in S0 and Sn). A path (A -> i; i -> B) has two
        turning point reactions. However, (A -> A + i; i -> B) must only have
        one turning point reaction, that corresponds to the formal basis S0 =
        {A} and Sn = {A, B}.  Since the states all contain formal species 
        (S0: {A}, S1: {A, i}, S2: {A, B}), only the reaction that guarantees a
        complete turnover of formal species is considered a turning point 
        (A -> A + i).

        Sidenote: Because waste species are treated as formal, they need to be
        produced after the turning point!
        """
        RFS = set() # Regular final states

        # Find last possible turning point j.
        # S0 --rxn1--> S1 --rxn2--> S2 ...
        # every Si < j is subset of S0, but Sj is not!
        states = self.states()
        S0 = self.minimal_initial_state
        for j, Si in enumerate(states):
            if not is_subset(self.formal(Si), S0):
                break

        T = Counter() # Final state
        for i in range(len(self), 0, -1):
            T |= self.formal(states[i])
            if i > j: 
                # Collect all formal species after turning point.
                continue
            # Get potential turning point reaction.
            [R, P] = self.cpath[i-1] # rxn 1 at pos 0
            remainder = self.formal(states[i-1] - R)
            if sum(remainder.values()) == 0:
                t = tuple(sorted(T.elements()))
                RFS.add(t)
        return frozenset(RFS)

    @property
    def signature(self):
        """ tuple: the signature of a path. """
        i = self.S0
        t = self.Sn
        w = self.width
        f = self.fc
        d = self.dfs
        r = self.rfs
        return tuple([i, t, w, f, d, r])

    def formal(self, S):
        """ Counter: multiset of formal species in the state S. """
        if self.fs is None:
            raise BasisFinderError('Path: no formal species given.')
        return Counter({k: v for k, v in S.items() if k in self.fs})

    def intermediate(self, S):
        """ Counter: multiset of non-formal species in the state S. """
        if self.fs is None:
            raise BasisFinderError('Path: no formal species given.')
        return Counter({k: v for k,v in S.items() if k not in self.fs})

    def states(self, counter = True):
        """ A list of states from S0 to Sn. 
        Args: counter (bool): returns states as Counters if True, returns lists otherwise.
        """
        if self._cstates is None:
            self._cstates = []
            self._cstates.append(self.minimal_initial_state)
            for [R, P] in self.cpath:
                self._cstates.append(self._cstates[-1] - R + P)
        if counter:
            # Typically, it is much more readable to work with counters.
            return self._cstates
        else:
            # Sometimes it is a *lot* faster to work with lists.
            return [sorted(s.elements()) for s in self._cstates]

    def _get_minits(self, path = None):
        """ Initialize internal S0 and Sn for this path object. """
        if path is None:
            path = self.cpath
        if self._minit is None:
            self._minit = Counter() # initial state
        if self._final is None:
            self._final = Counter() # current state
        for [R, P] in path:
            self._minit += (R - self._final)
            self._final = (self._final - R) + P

    def __add__(self, other):
        """ Path: the concatenation of two Path objects. """
        assert type(self) == type(other)
        combo = Path(self.lpath + other.lpath)
        combo.fs = self.fs # force copy?

        # set by: self._get_minits()
        if self._minit:
            combo._minit = self._minit + Counter()
            combo._final = self._final + Counter()
            combo._get_minits(path = other.cpath)
        # set by: decomposed_final_states
        if combo.is_semiformal and self._dfinals is not None:
            combo._dfinals = self._dfinals
            combo.decomposed_final_states(path = other.cpath)
        return combo

    def __len__(self):
        return len(self.lpath)

    def __eq__(self, other):
        return self.signature == other.signature

    def __ne__(self, other):
        return not self == other

    def __hash__(self):
        return hash(self.signature)

    def __str__(self):
        return '{}'.format('; '.join(self.pretty_path))

    def __repr__(self):
        return "Path(l={}, {})".format(len(self), self.signature)

def tidy(queue, crn, fs, TC = None, bound = None):
    """ BFS to test if a pathway p has a strong closing pathway q.

    Typically, queue contains only one state: all intermediate species of the
    final state T of a pathay p. The function attempts to find a closing
    pathway q, which produces a formal state (without requiring formal species
    as reactants).

    Intended context & justification: The bound for tidyness checks is not
    specifically mentioned in the paper (STW2020), but it follows from
    Corollary 4.3 and Theorem 4.8 that there must not be a semiformal path 
    p + q with width > w, once it is known that there is no semiformal path
    between w and w_max+1.
    Hence, the present implementation:
        - tidy(...) is called with the current width bound w_max whenever a new
          undecomposable pathway has been found.
        - if tidy(...) returns False, the algorithm terminates.
        - TC holds the results for each tested final state:
            * TC[state] = True # CRN is tidy from that state 
            * TC[state] = queue # Pick up up the search here in case the width
              bound changes.
        - After it has been established that there are no undecomposable
          (strictly) semiformal paths with width >= w, every state in TC where
          TC[state] != True is checked for tidyness again with bound w. If
          still no formal state is found then the CRN is not tidy.

    Args:
        queue (list): A list of states (states must contain only intermediate species).
        crn (list): The CRN.
        fs (set): A set of formal species.
        TC (dict, optional): A Tidy-Check dictionary that cotains known tidy states.
        bound (int, optional): A width bound to limit the search for a pathway q.

    Returns:
        True: if a strong closing pathway exists
        False: if no strong closing pathway exists
        queue: a list of states that exceed the current bound
    """
 
    def next_state(s, rxn):
        # Applies reaction rxn to state s.
        s = list(s)
        for x in rxn[0]:
            s.remove(x)
        s += rxn[1]
        return s

    # Keep this. queue should not be a single state but a list of states.
    assert all(tuple(intermediate(x, fs)) == x for x in queue)

    mem = queue[:]
    while len(queue) > 0:
        queue.sort(key = len)
        iS = queue[0] 
        if len(iS) == 0:
            return True
        elif TC and TC.get(iS, False) is True: 
            return True
        elif bound and len(iS) > bound:
            return queue
        queue = queue[1:]
        for rxn in crn:
            try:
                nS = tuple(sorted(intermediate(next_state(iS, rxn), fs)))
                if nS not in mem: 
                    mem.append(nS)
                    queue.append(nS)
            except ValueError as err:
                # iS did not contain the required reactants.
                continue
    return False

def crn_properties(crn, fs):
    """ Get different CRN properties.

    Returns:
        b (int): the branching factor
        iR: max over |intermediate(R)| of (R,P) in crn
        fRiR: set of (|formal(R)|, |intermediate(R)|)
        nw: nonwaste species for linear check
    """
    # NOTE: Def 27 of STW uses "nonwaste" species to define monomolecular
    # substructure, but nonwastes = intermediates, because wastes must
    # be treated as formal species.
    nw, w, _ = assign_crn_species(crn, fs)
    assert len(nw & w) == 0 # assert that we can use intermediates as non-wastes.
    linear = True
    for [r, p] in crn:
        r1 = [x for x in r if x in nw]
        p1 = [x for x in p if x in nw]
        if len(r1) > 1 or len(p1) > 1: 
            linear = False
            break
    log.debug('Monomolecular substructure: {}'.format(linear))
    b  = max(max(len(r),len(p)) for [r, p] in crn)
    iR = max(len(intermediate(r, fs)) for [r, p] in crn) 
    fRiR = set((len(formal(r, fs)), len(intermediate(r, fs))) for [r, p] in crn)
    log.debug('Branching Factors: b = {}, iR = {}, fRiR = {}'.format(b, iR, fRiR))
    return b, iR, fRiR, nw if linear else None

def enumerate_basis(crn, fs, inter = None):
    """ Find the formal basis of a CRN.

    Args:
        crn (List[[R], [P]]): A list of irreversible reactions. First a list of
            reactants, second a list of products.
        fs (set(str)): A set of formal species.
        interpretation (dict, optional): An interpretation dictionary for the 
            "integrated hybrid" mode. Defaults to None.

    Returns:
        basis_raw, basis_int: The formal basis found without and with an 
            interpretation dictionary. 
    """
    TidyCheck = dict() # A cache to remember which final states are Tidy. 
    b, b_r, fRiR, nonw = crn_properties(crn, fs)

    # Uses some constants: w_max, i_max, nonw, fs, but not crn (in case we want it random)
    def enumerate_elementary_basis(p, crn):
        """Enumerates all signatures and thereby the relevant paths.

        Note that the search is bounded by the external constants w_max, i_max, and that
        it relies on the definition of formal species (fs) and non-wastes (nonw).
        """
        if not p.is_semiformal:
            return
        if p.width > w_max: 
            return 
        if sum(p.minimal_initial_state.values()) > i_max: 
            return
        if nonw is not None and not p.is_linear(nonw):
            return
        for df in p.dfs:
            if is_formal_state(df, fs):
                return

        sig = p.signature
        if sig in signatures:
            return
        signatures.add(sig)

        (i, fin, w, fc, dfs, rfs) = sig
        if len(dfs) == 0: # Ok, we found a valid path? Check if the system is tidy and regular!
            log.debug(f"New undecomposable path: {p}")
            ifin = tuple(sorted(intermediate(fin, fs)))
            if ifin not in TidyCheck or TidyCheck[ifin] is not True:
                log.debug(f"Checking if tidy from state: {ifin} ...")
                queue = [ifin] if ifin not in TidyCheck else TidyCheck[ifin]
                TidyCheck[ifin] = tidy(queue, crn, fs, TidyCheck, bound = w_max)
                if TidyCheck[ifin] is False:
                    log.info("The given system is not tidy.")
                    log.info(" - from initial state: {}".format(p.S0))
                    [log.info('    {}'.format(r)) for r in p.pretty_path]
                    raise NoFormalBasisError("The given system is not tidy.")
                elif TidyCheck[ifin] is not True:
                    log.debug(f"The system is not tidy from initial state {p.S0} and final state {p.Sn} given the current width bound: {w_max}.")
                log.debug(f"... done.")
            if len(p) > 0 and p.is_formal:
                if inter: # integrated hybrid theory
                    fs1 = set(interpret(fs, inter))
                    p1 = [[interpret(R, inter), interpret(P, inter)] for [R, P] in p.lpath]
                    p1 = Path(p1, fs1)
                    i1 = p1.S0
                    t1 = p1.Sn
                    if t1 not in p1.rfs:
                        log.info("The given system is not regular.")
                        log.info(" - from initial state: {}".format(p.S0))
                        [log.info('    {}'.format(r)) for r in p1.pretty_path]
                        raise NoFormalBasisError("The given system is not regular.")
                elif fin not in rfs:
                    log.info("The given system is not regular.")
                    log.info(" - from initial state: {}".format(p.S0))
                    [log.info('    {}'.format(r)) for r in p.pretty_path]
                    raise NoFormalBasisError("The given system is not regular.")
                ebasis.append(p)
                return # adding reaction must yield a strongly decomposable pathway.

        perm = False # turn on if you want a randomized/shuffled search
        if perm:
            import copy, random
            crn_perm = copy.copy(crn) 
            random.shuffle(crn_perm)
            tmp = crn_perm
        else :
            tmp = crn
        for r in crn:
            enumerate_elementary_basis(p + Path([r]), tmp)

    ############################################################################
    #          On w and w_max as criteria to stop enumeration:                 #
    # By STW2019 Theorem 4.2, the width of an undecomposable semiformal path   #
    # w_p and the width of that path after removing the last reaction w_{p-1}  #
    # are related by:                                                          #
    #                       w_p >= w_{p-1} >= w_p - b                          #
    # Also, p-1 can be decomposed into at most b_r undecomposable pathways     #
    # (see Thm 4.10), and at least one of those pathways must have width:      #
    #                           w >= (w-b)/b_r                                 #
    # Now take an interval between w and w_max+1, where                        #
    #                           w_max = w * b_r + b,                           #
    # then any undecomposable semiformal pathway with width v > w_max can be   #
    # decomposed into at least one semiformal path that has width in this      #
    # interval or (if it has w >= wmax) it can be further decomposed to have   #
    # width in this interval.                                                  #
    # Conversely, if no undecomosable semiformal path exists in that interval, #
    # there exists no semiformal path with width greater than w_max.           #
    ############################################################################

    w_max, i_max = 0, 0
    while True:
        log.debug("Current bounds: w_max = {}, i_max = {}".format(w_max, i_max))
        signatures, ebasis = set(), [] # We actually enumerate signatures, but care about paths.
        enumerate_elementary_basis(Path([], fs), crn)
        log.debug('After enumation of pathways: len(basis) = {}'.format(len(ebasis)))
        [log.debug(f'   [{" + ".join(p.S0)} -> {" + ".join(p.Sn)}]  {p}') for p in ebasis]
        new_w, new_i = 0, 0
        for (i, f, w, fc, dfs, rfs) in signatures:
            # NEW: only *strongly* semiformal paths (no formal paths)
            if len(dfs) == 0 and not is_formal_state(f, fs):
                log.debug(f'w: {w}, S0: {i}, Sn: {f}') 
                new_w = max(new_w, w)
                new_i = max(new_i, len(i)) # Size of initial (formal) state.
        w_t = new_w * b_r + b
        i_t = 0
        for (fr, ir) in fRiR:
            i_t = max(i_t, new_i * ir + fr)
        if w_t <= w_max and i_t <= i_max:
            for ifin in TidyCheck:
                if TidyCheck[ifin] is not True:
                    if tidy(TidyCheck[ifin], crn, fs, TidyCheck, bound = new_w) is not True:
                        raise NoFormalBasisError(f"The given system is not tidy from state {ifin}.")
            break
        w_max = w_t
        i_max = i_t

    # Now translate elementary basis into formal basis:
    fbasis_raw = []
    fbasis_int = [] # interpretation
    for path in ebasis:
        S = path.minimal_initial_state
        T = path.final_state
        r = [sorted(S.elements()), sorted(T.elements())] 
        fbasis_raw.append(r)

        if inter: # integrated hybrid theory
            fs1 = set(interpret(fs, inter))
            p1 = [[interpret(R, inter), interpret(P, inter)] for [R, P] in path.lpath]
            p1 = Path(p1, fs1)
            S1 = p1.minimal_initial_state
            T1 = p1.final_state
            r = [sorted(S1.elements()), sorted(T1.elements())] 
            fbasis_int.append(r)
    return fbasis_raw, fbasis_int

def get_crn_modules(crn, intermediates):
    """ Partition the CRN into modules that do not share intermediate species. """
    def ancestor(x):
        if parent[x] != x:
            parent[x] = ancestor(parent[x])
        return parent[x]

    # Get all the divisions
    parent   = {i: i  for i in intermediates}
    division = {i: [] for i in intermediates}

    # First, determine which intermediates occur together...
    for rxn in crn:
        t = [x for x in rxn[0] + rxn[1] if x in intermediates]
        if len(t) > 1:
            # parent of all intermediates should point to the same "ancestor".
            z = ancestor(t[0])
            for x in t[1:]:
                y = ancestor(x)
                if z != y:
                    parent[y] = z
    i = 0 # TODO: test this function!
    for rxn in crn:
        t = [x for x in rxn[0] + rxn[1] if x in intermediates]
        if len(t) > 0:
            z = ancestor(t[0])
            division[z].append(rxn)
        else: # formals, wastes only
            division[i] = [rxn]
            i += 1
    return [v for v in division.values() if len(v) > 0]

def get_formal_basis(crn, fs, modular = True, interpretation = None):
    """Finds all formal reactions in a CRN.
    
    STW2019 - Def13: The set of prime pathways in a given CRN is called the
    *elementary basis* of the CRN. The *formal basis* is the set of (initial
    state, final state) pairs of the pathways in the elementatry basis.

    Args:
      crn (List[[R], [P]]): A list of irreversible reactions. First a list of
            reactants, second a list of products.
      fs (set(str)): A set of formal species.
      modular (bool, optional): Chop the CRN into modules and then find the basis
            separately for each of those modules. Defaults to True.
      interpretation (dict, optional): An interpretation dictionary for the 
            "integrated hybrid" mode. Defaults to None.

    Returns:
        basis_raw, basis_int: The formal basis found without and with an 
            interpretation dictionary. 
    """

    intermediates, wastes, reactive_waste = assign_crn_species(crn, fs)
    if len(wastes):
        if len(reactive_waste):
            log.warning('Reactive waste species detected: ' + \
                        f'{{{", ".join(natural_sort(reactive_waste))}}}')
        log.info(f'{len(wastes)} waste species are treated as formal: ' + \
                 f'{{{", ".join(natural_sort(wastes))}}}')
        fs |= wastes
 
    if modular:
        divs = sorted(get_crn_modules(crn, intermediates), key = lambda x: len(x))
        log.info(f"Divided the implementation CRN into {len(divs)} modules " + \
                f"with {[len(d) for d in divs]} reactions.")
        basis_raw = []
        basis_int = []
        for e, mod in enumerate(divs, 1):
            log.info("Verifying module {}:".format(e))
            [log.info(f'    {r}') for r in pretty_crn(mod)]
            b_raw, b_int = enumerate_basis(mod, fs, inter = interpretation)
            log.info("Formal basis of the current module:")
            [log.info('    {}'.format(r)) for r in pretty_crn(b_raw)]
            if interpretation is not None:
                log.info("Formal basis of the current module (integrated hybrid):")
                [log.info('    {}'.format(r)) for r in pretty_crn(b_int)]
            log.info('')
            basis_raw += b_raw
            basis_int += b_int
    else:
        basis_raw, basis_int = enumerate_basis(crn, fs, inter = interpretation)

    return clean_crn(basis_raw), clean_crn(basis_int)

def pathway_decomposition_eq(crns, fs, modular = True):
    """ Test two CRNs for pathway equivalence.

    Args:
        crns: List of one or more CRNs.
        fs: Set of formal species.
    Returns:
        True: if formal basis of all crns is equivalent
        False: otherwise
    """
    v, last_basis = None, None
    inter = None
    for e, crn in enumerate(crns, 1):
        _, wastes, reactive_waste = assign_crn_species(crn, fs)
        if len(reactive_waste):
            log.warning('Reactive waste species detected: ' + \
                        f'{{{", ".join(natural_sort(reactive_waste))}}}')
        if len(wastes):
            log.info(f'{len(wastes)} waste species are treated as formal: ' + \
                     f'{{{", ".join(natural_sort(wastes))}}}')
            inter = {w: [] for w in wastes}
        try:
            fbasis, _ = get_formal_basis(crn, fs | wastes, modular = modular)
        except NoFormalBasisError as err:
            log.info("Could not find formal basis: {}".format(err))
            return False

        log.info('Formal basis {}:\n  {}'.format(e, "\n  ".join(pretty_crn(fbasis))))
        if inter is not None:
            fbasis = clean_crn(fbasis, inter = inter)
        if last_basis is not None:
            v = sorted(last_basis) == sorted(fbasis)
            if v is False:
                log.info(f"The CRNs {e-1} and {e} are not pathway-decomposition equivalent.")
                return False
        last_basis = fbasis
    return True




"""
Parser for formal Chemical Reaction Networks.

Note: This file was adapted from the python package: `Nuskell`, which is
    developed at the *DNA and Natural Algorithms Group*, Caltech.  
    Original code written by Seung Woo Shin (seungwoo.theory@gmail.com).

Test using tests/test_crn_parser.py.
"""

from pyparsing import (Word, Literal, Group, Suppress, Combine, Optional, ParseException,
                       alphas, nums, alphanums, delimitedList, StringStart, StringEnd, LineEnd,
                       ZeroOrMore, OneOrMore, pythonStyleComment, ParseElementEnhance)

class CRNParseError(Exception):
    pass

def crn_document_setup(modular = False):
    """Parse a formal chemical reaction network.

    Args:
      modular <optional:bool>: Adds an additional nesting for modules within a
        CRN. Use one line per module (';' separates reactions).

    Format:
      # A list of reactions, optionally with reaction rates:
      # <- this is a comment!
      B + B -> C    # [k = 1]
      C + A <=> D   # [kf = 1, kr = 1]
      <=> A  [kf = 15, kr = 6]

      # Note that you can write multiple reactions in one line:
      A + 2C -> E [k = 13.78]; E + F <=> 2A  [kf = 13, kr = 14]

    Returns:

    """
    # NOTE: If you want to add support for multiple modules per line, you can use
    # the '|' character.

    W = Word
    G = Group
    S = Suppress
    O = Optional
    C = Combine
    L = Literal

    def T(x, tag):
        """ Return a *Tag* to distinguish (ir)reversible reactions """
        def TPA(tag):
            return lambda s, l, t: [tag] + t.asList()
        return x.setParseAction(TPA(tag))

    crn_DWC = "".join(
        [x for x in ParseElementEnhance.DEFAULT_WHITE_CHARS if x != "\n"])
    ParseElementEnhance.setDefaultWhitespaceChars(crn_DWC)

    identifier = W(alphas, alphanums + "_")

    multiplier = W(nums)
    species = G(O(multiplier) + identifier)

    number = W(nums, nums)
    num_flt = C(number + O(L('.') + number))
    num_sci = C(number + O(L('.') + number) + L('e') + O(L('-') | L('+')) + W(nums))
    gorf = num_sci | num_flt

    # Make specification of forward, backward, reverse more flexible
    kf = S('kf') | S('fw')
    kr = S('kr') | S('bw') | S('rv')

    k = G(S('[') + O(S('k') + S('=')) + gorf + S(']'))
    rev_k = G(S('[') + kf + S('=') + gorf + S(',') + kr + S('=') + gorf + S(']')) | G(S('[') + gorf + S(',') + gorf + S(']'))

    concentration = T(species + S('@') + G(L("initial") | L("i") | L("constant") | L("c")) + G(gorf), 'concentration')

    reaction = T(G(O(delimitedList(species, "+"))) +
                 S("->") +
                 G(O(delimitedList(species, "+"))) + O(k), 'irreversible')

    rev_reaction = T(G(O(delimitedList(species, "+"))) +
                     S("<=>") +
                     G(O(delimitedList(species, "+"))) + O(rev_k), 'reversible')

    expr = G(reaction | rev_reaction | concentration)

    if modular:
        module = G(expr + ZeroOrMore(S(";") + expr))
    else:
        module = expr + ZeroOrMore(S(";") + expr)

    crn = OneOrMore(module + ZeroOrMore(S(LineEnd())))

    document = StringStart() + ZeroOrMore(S(LineEnd())) + crn + StringEnd()
    document.ignore(pythonStyleComment)
    return document

def flint(inp):
    try:
        return int(inp) if float(inp) == int(float(inp)) else float(inp)
    except OverflowError:
        return inp

def post_process(crn, modular = False, defaultrate = 1, defaultmode = 'initial', defaultconc = 0):
    """Process a parsed CRN.

    Does:
     - remove 'reversible' and 'irreversible' tags.
        (this information is stored implicitly with one or two rates).
     - Translates species multipliers to multisets
     - Extracts the set of all species in the CRN.
     - Adds 'None' as reaction rate if no rate was specified.

    Does not:
     - Check if reactions are specified twice
     - Check stochiometry

    Returns:
        new: the CRN in format [[r],[p],k]
        species: dictionary species[name]=('initial', float)

    """
    def remove_multipliers(species):
        flat = []
        for s in species:
            if len(s) == 1:
                flat.append(s[0])
            elif len(s) == 2:
                ss = [s[1]] * int(s[0])
                flat.extend(ss)
        return flat

    default = dict()
    species = dict()
    def get_module(crn):
        new = []
        for line in crn:
            if line[0] == 'concentration':
                spe = line[1][0]
                ini = 'initial' if line[2][0][0] == 'i' else 'constant'
                num = line[3][0]
                if spe in species:
                    raise CRNParseError(f'Duplicate concentration specification for {spe}!')
                species[spe] = (ini, flint(num))
                continue
            elif len(line) == 3:
                # No rate specified
                t, r, p = line
                r = remove_multipliers(r)
                p = remove_multipliers(p)
                if t == 'reversible':
                    new.append([r, p, [defaultrate, defaultrate]])
                elif t == 'irreversible':
                    new.append([r, p, [defaultrate]])
                else:
                    raise CRNParseError('Wrong CRN format!')
            elif len(line) == 4:
                t, r, p, k = line
                r = remove_multipliers(r)
                p = remove_multipliers(p)
                k = [flint(v) for v in k]
                if t == 'reversible':
                    assert len(k) == 2
                    new.append([r, p, k])
                elif t == 'irreversible':
                    assert len(k) == 1
                    new.append([r, p, k])
                else:
                    raise CRNParseError('Wrong CRN format!')
            else:
                raise CRNParseError('Wrong CRN format!')
            for s in r + p:
                if s not in default:
                    default[s] = (defaultmode, defaultconc)
        return new

    if modular:
        new = []
        for module in crn:
            mcrn = get_module(module)
            if mcrn:
                new.append(mcrn)
    else:
        new = get_module(crn)

    default.update(species)
    return new, default

def parse_crn_file(filename, process = True, modular = False, **kwargs):
    """Parses a CRN from a file.

    Args:
      filename (<str>): The name of the file for parsing.
      process (optional <bool>): Process the format of the returned CRN.

    Returns:
      crn (<lol>): List of list representation of a CRN
      species (<set()>): A set of all involved species (only when process=True)

    """
    crn_document = crn_document_setup(modular)
    if process:
        return post_process(crn_document.parseFile(filename, parseAll=True).asList(), modular, **kwargs)
    else:
        return crn_document.parseFile(filename, parseAll=True).asList()

def parse_crn_string(data, process = True, modular = False, **kwargs):
    """Parses a CRN from a string.

    Args:
      filename (<str>): The name of the file for parsing.
      process (optional <bool>): Process the format of the returned CRN.

    Returns:
      crn (<lol>): List of list representation of a CRN
      species (<set()>): A set of all involved species (only when process=True)

    """
    crn_document = crn_document_setup(modular)
    if process:
        return post_process(crn_document.parseString(data).asList(), modular, **kwargs)
    else:
        return crn_document.parseString(data).asList()





from collections import Counter
from itertools import product, permutations

# helper functions copied from crn_bisimulation.py
def msleq(x,y):
    # True if (multisets) x <= y (vector comparison)
    for k in x:
        if x[k] > y[k]:
            return False
    return True

def mstimes(s, l):
    # return l*s for integer l, multiset s
    c = Counter()
    for k in s:
        c[k] = l * s[k]
    return c

def interpret(s, intrp):
    # return interpretation of s according to intrp
    ss = s.copy()
    for k in list(s):
        if k in intrp:
            v = ss.pop(k)
            for i in range(v):
                ss += intrp[k]
    return ss

def interleq(x, y, intrp):
    # True if m(x) <= m(y) with m given by interpretation intrp
    return msleq(interpret(x,intrp),interpret(y,intrp))
def solve_contejean_devie(a):
    """ Algorithm from Contejean & Devie 1994.

    Find a non-negative and non-trivial integer solution x of the equation ax=0.
    Return [] when there is no such solution.
    """
    q = len(a[0])
    
    def multi(x, y):
        s = 0
        for i in range(len(x)):
            s = s + x[i] * y[i]
        return s
    
    def sub(x):
        s = []
        for i in range(len(a)):
            s.append(multi(a[i],x))
        return s
    
    def Min(b,t):
        if not b:
            return True
        else:
            for i in range(len(b)):
                r = True;
                for j in range(q):
                    r = r and (b[i][j] <= t[j])
                if r:
                    return False
            return True
        
    e = []
    for i in range(q):
        e.append([])
        for j in range(len(a)):
            e[i].append(a[j][i])     
    p = []
    frozen = []
    for i in range(q):
        p.append([1 if j == i else 0 for j in range(q)])
        frozen.append([i == q-1 or j < i for j in range(q)])
    zero = [0 for i in range(len(a))]
    zero1 = [0 for i in range(q)]
    b = []
    while p:
        t = p.pop()
        if sub(t) == zero:
            if t[q-1] == 1:
                return t    # just get the first solution, not all solutions (unlike C&D 1994).
            b.append(list(t))
            frozen.pop()
        else:
            f = frozen.pop()
            for i in range(q):
                if not f[i] and (multi(sub(t), e[i]) < 0):
                    tmp = list(t)
                    tmp[i] += 1
                    if Min(b, tmp):
                        if i == q-1:
                            f[i] = True
                        p.append(tmp)
                        frozen.append(list(f))
                    f[i] = True
    return []

def solve(a):
    # wrapper method to solve a system of equations with method of choice
    return solve_contejean_devie(a)



# DEPRECATED STATS HERE
def find_one_trivial(fcrn, icrn, fs, intrp, permcheck = 'graphsearch'):

    # All unknown implementation reactions (i.e. those with some unknown
    # species) must be trivial.  Build the matrix for the the "solve" function,
    # to see whether the interpretation can be completed as required.  Also
    # note that "unknow" is not used; the unknown species are recalculated here
    # because "unknow" contains only those implementation reactions that have
    # not been solved by the row search, but other rows (i.e. implementation
    # reactions) may be implicitly solved by the partial interpretation.
    sicrn = subst(icrn, intrp)

    # List of implementation reactions with an unknown species.
    unknown = [i for i, irxn in enumerate(sicrn) if len(set(irxn[0]) - fs) or \
                                                   len(set(irxn[1]) - fs)]
    # All species that remain unknown in current (partial) interpretation.
    unassigned = set(s for irxn in sicrn for s in set(irxn[0]) - fs | set(irxn[1]) - fs)
    
    # Find species that do not satisfy atomic condition
    atoms_needed = list(fs - set().union(*[set(a) for a in intrp.values() if sum(a.values()) == 1]))

    log.info(f'Trivial reaction solver: {unknown=}')
    log.info(f'Missing atoms: {atoms_needed}')
    log.info(f'Unknown species: {unassigned}')

    for assign in permutations(unassigned, len(atoms_needed)): # works even if l == 0
        # each assign is a tuple of implementation species to be interpreted as
        # exactly one formal species, matching the order of atoms_needed.
        log.debug(f'{assign=}')

        parti = intrp.copy()
        for i, a in enumerate(atoms_needed):
            assert assign[i] not in parti
            parti[assign[i]] = Counter({a: 1})
        log.debug(f'{parti=}')

        sicrn = subst(icrn, parti)
        T = updateT(fcrn, sicrn, fs)
        if (not checkT(T)) or any([T[i][-1] is False for i in unknown]):
            continue

        ulist = [sp for sp in unassigned if sp not in assign]
        log.debug(f'{ulist=}')
        if not ulist:
            out, info = check_permissive(fcrn, icrn, fs, parti, permcheck)
            if out:
                yield parti
            continue

        # A mini table. 
        # For every unknown implementation reaction make a list:
        #   - for each unassigend species store #ur - #up
        a = []
        for i in unknown:
            irxn = sicrn[i]
            log.debug(irxn)
            # A list of #ur - #up
            a.append([irxn[0][u]-irxn[1][u] for u in ulist] + [0])
        log.debug(f'{a=}')

        # prepare for adding species to the counter ...
        for u in ulist:
            parti[u] = Counter()
        log.debug(f'{parti=}')

        check = True
        for fsp in fs:
            log.warning(f'{fsp=}')
            for e, i in enumerate(unknown):
                irxn = sicrn[i]
                # A list of #fsr - #fsp
                a[e][-1] = irxn[0][fsp]-irxn[1][fsp]
            log.debug(f'{a=}')

            # Alright, this can find one, but only one solution... 
            s = solve(a)
            log.debug(f'{s=}')
            if s == []:
                check = False
                break
            else:
                for j, u in enumerate(ulist):
                    parti[u][fsp] = s[j]
        if check:
            for isp in ulist:
                parti[isp] = parti[isp] + Counter()
            out, info = check_permissive(fcrn, icrn, fs, parti, permcheck)
            if out:
                yield parti
    return

def moduleCond(module, formCommon, implCommon, intrp):
    """
    # check whether the modularity condition (every implementation species can
    # turn into common species with the same interpretation) is satisfied
    # assumes intrp is complete and filtered, so intrp.keys() is a list of all
    # implementation species in this module (and no others) algorithm is
    # basically the graphsearch algorithm from perm, where all "minimal states"
    # are each exactly one implementation species

    # canBreak[k] is:
    #   True if species k is known to decompose (via trivial reactions) into
    #     species which are each either in implCommon or have an interpretation
    #     containing nothing in formCommon, or known to decompose as such if
    #     some set of other species each with interpretation strictly < do
    #   [reach,produce] where reach is set of implementation species with
    #     interpretation equal to that of k (known) reachable from k, and
    #     produce is set of null species producible in a loop from k to k
    """
    module = [[Counter(part) for part in rxn] for rxn in module]
    intrp = {k : Counter(v) for k, v in intrp.items()}

    canBreak = {k: ((k in implCommon) or set(intrp[k]).isdisjoint(formCommon)
                    or [set(),set()])
                for k in intrp}
    changed = True

    tr = [rxn for rxn in module if (lambda x,y: msleq(x,y) and msleq(y,x))
          (interpret(rxn[0],intrp),interpret(rxn[1],intrp))]

    while changed:
        changed = False
        for k in canBreak:
            if canBreak[k] is True: continue

            for rxn in tr:
                if k in rxn[0] and \
                   set(rxn[0] - Counter([k])).issubset(canBreak[k][1]):
                    # reactants of rxn are one copy of k and some null species
                    #  producible in a loop from k to k
                    nulls = set()
                    theOne = None
                    for sp in rxn[1]:
                        if intrp[sp] == Counter():
                            nulls.add(sp)
                        elif not msleq(intrp[k],intrp[sp]):
                            canBreak[k] = True
                            changed = True
                            break
                        else:
                            if canBreak[sp] is True:
                                canBreak[k] = True
                                changed = True
                                break
                            theOne = sp

                    if canBreak[k] is True:
                        break

                    if theOne not in canBreak[k][0]:
                        canBreak[k][0].add(theOne)
                        changed = True

                    if not (canBreak[theOne][0] <= canBreak[k][0]):
                        canBreak[k][0] |= canBreak[theOne][0]
                        changed = True

                    if k in canBreak[theOne][0]:
                        loopable = nulls | canBreak[theOne][1]
                        if not loopable <= canBreak[k][1]:
                            canBreak[k][1] |= loopable
                            changed = True

    return all([canBreak[k] is True for k in canBreak])

def update(fcrn, icrn, fs):
    # the um, er, um, completely recalculates the table from scratch.
    # assumes subst has already been applied to implementation icrn.
    # This should be logically equivalent to the UpdateTable in the MS thesis
    # for compiled DNA reactions (we think).

    # WARNING:
    # If an implementation CRN has directly catalytic species, the code below
    # may fail (though thesis psuedocode is correct).  
    # E.g.  i3 + i7 --> i12 + i7 + i8
    m = len(icrn)
    r = []
    for i in range(len(icrn)):
        rr = []
        for j in range(len(fcrn)):
            t1 = fcrn[j][0] - icrn[i][0]
            t2 = icrn[i][0] - fcrn[j][0]
            t3 = fcrn[j][1] - icrn[i][1]
            t4 = icrn[i][1] - fcrn[j][1]
            if set(t2).isdisjoint(set(fs)) and set(t4).isdisjoint(set(fs)):
                if list(t1.keys()) == []:
                    if list(t3.keys()) == []:
                        rr.append(True)
                    else:
                        if list(t4.keys()) == []:
                            rr.append(False)
                        else:
                            rr.append(True)
                else:
                    if list(t2.keys()) == []:
                        rr.append(False)
                    else:
                        if list(t3.keys()) == []:
                            rr.append(True)
                        else:
                            if list(t4.keys()) == []:
                                rr.append(False)
                            else:
                                rr.append(True)
            else:
                rr.append(False)
        t1 = icrn[i][0] - icrn[i][1]
        t2 = icrn[i][1] - icrn[i][0]
        if (set(t1).isdisjoint(set(fs)) or not set(t2).issubset(set(fs))) and \
                (set(t2).isdisjoint(set(fs)) or not set(t1).issubset(set(fs))):
            rr.append(True)
        else:
            rr.append(False)
        r.append(rr)
    return r

def subsets(x):
    # generates all subsets of multiset x
    [ks, vs] = zip(*x.items()) if len(x) else [[],[]]
    vs = [list(range(v + 1)) for v in vs]
    for prod in product(*vs):
        # calls to keys and values with no dictionary modifications in between
        #  should produce the keys and values in the same order,
        #  and product should respect that order (I think)
        yield Counter(dict(list(zip(ks, prod)))) + Counter()

def enum(n, s, weights = None):
    """Partition multiset s into n ordered (possibly empty) parts.  

    For example:
        enum(2, [a, b]) = [ [[], [a, b]], 
                            [[a], [b]], 
                            [[b], [a]], 
                            [[a, b],[]] ]

    If weights are given, enumerates all lists l of n multisets such that 
        s = sum(weights[i] * l[i]) 
    (if weights are not given, equivalent to weights[i] = 1 for all i)
    """
    if weights is None:
        weights = [1] * n

    if n == 0:
        yield []
        return

    def msdiv(s, l):
        # return s/l for integer l, multiset s if l divides s, otherwise False
        # l divides s if l divides s[k] for every key k
        c = Counter()
        for k in s:
            c[k] = int(s[k]/l)
            if c[k] * l != s[k]:
                return False
        return c

    if len(weights) < n:
        raise IndexError(f'{len(weights)} weights given for {n} parts.')
    elif weights[0] < 0:
        raise ValueError('Negative weight given.')
    elif weights[0] == 0:
        for j in enum(n-1, s, weights[1:]):
            yield [Counter()] + j
        return

    if n == 1:
        sdivw = msdiv(s, weights[0])
        if sdivw is not False: 
            yield [sdivw]
        return

    for i in subsets(s):
        ss = mstimes(i, weights[0])
        if not msleq(ss, s):
            continue
        for j in enum(n-1, s-ss):
            yield [i] + j

def searchc(fcrn, icrn, fs, intrp, unknown, depth, permcheck, state):
    # Search column.  I.e. make sure every formal reaction can be implemented.
    """
    fcrn, icrn, fs, unknown and intrp, permcheck remain constant, i think.

    Depth will update, state will update.

    """
    from .crn_bisimulation import subst, updateT, checkT, enumL

    # A quick check if the delimiting condition is still satisfied.
    sicrn = subst(icrn, intrp)
    T = updateT(fcrn, sicrn, fs)
    if not checkT(T):
        yield False
        yield state
        return

    # Rollback to original interpretation if you reach the max_depth,
    # I assume this is a hack to do some sort of depth/breadth first search.
    intr, max_depth = state[0:2]
    if max_depth >= 0 and depth > max_depth:
        intr = intrp.copy()
        max_depth = depth

    found = False # Changes to True if you found something ... but what?

    # Find the next column to solve. 
    c, worst = -1, len(icrn) + 1
    for i in unknown: # The column indices of formal reactions 
        # Count the number of True's in the column.
        tmp = sum(T[j][i] for j in range(len(icrn)))
        if tmp < worst:
            c, worst = i, tmp

    if c < 0: # Done with column search, transition to row search.
        untmp = []
        for i in range(len(icrn)):
            if not (set(sicrn[i][0])-set(fs) == set([]) and \
                    set(sicrn[i][1])-set(fs) == set([])):
                untmp.append(i)
        out = searchr(fcrn, icrn, fs, untmp, intrp, depth, permcheck,
                      [intr, max_depth] + state[2:])
        if next(out):
            yield True
            found = True
            for outintrp in out:
                yield outintrp
        else:
            yield False
            yield next(out)
            return
    else:
        # search_column(T, c, unknown, icrn, fcrn, intrp, sicrn)
        # This part will recursively call searchc 

        untmp = list(unknown)
        untmp.remove(c)
        for k in range(len(icrn)):
            if not T[k][c]:
                continue
            # left 
            ul = sicrn[k][0] - fcrn[c][0]
            sl = fcrn[c][0] - sicrn[k][0]
            [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
            tmpl = enumL(len(ul), list(sl.elements()), vl)
            # right
            ur = sicrn[k][1] - fcrn[c][1]
            sr = fcrn[c][1] - sicrn[k][1]
            [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
            tmpr = enumL(len(ur), list(sr.elements()), vr)

            for i, j in product(tmpl, tmpr):
                intrpleft = dict(zip(kl, i))
                intrpright = dict(zip(kr, j))
                for key in set(intrpleft) & set(intrpright):
                    if any([sorted(intrpleft[key][fsp]) != sorted(intrpright[key][fsp]) \
                        for fsp in fs]):
                        # Incompatible dictionaries!
                        break
                else:
                    itmp = intrp.copy()
                    itmp.update({k: Counter(v) for k, v in intrpleft.items()})
                    itmp.update({k: Counter(v) for k, v in intrpright.items()})
                    out = searchc(fcrn, icrn, fs, 
                                  itmp, 
                                  untmp, 
                                  depth + 1,
                                  permcheck, 
                                  [intr, max_depth] + state[2:])
                    if next(out):
                        if not found:
                            found = True
                            yield True
                        for outintrp in out:
                            yield outintrp
                    else:
                        state = next(out)
                        intr, max_depth = state[0:2]

    if not found:
        yield False
        yield [intr, max_depth] + state[2:]

def equations(fcrn, icrn, fs, intrp, permcheck, state):
    # All unknown implementation reactions (i.e. those with some unknown
    # species) must be trivial.  Build the matrix for the the "solve" function,
    # to see whether the interpretation can be completed as required.  Also
    # note that "unknow" is not used; the unknown species are recalculated here
    # because "unknow" contains only those implementation reactions that have
    # not been solved by the row search, but other rows (i.e. implementation
    # reactions) may be implicitly solved by the partial interpretation.
    unknown = []
    ust = set([])
    sicrn = subst(icrn, intrp)
    for i in range(len(sicrn)):
        tmp = set(sicrn[i][0])-set(fs)
        tmp |= set(sicrn[i][1])-set(fs)
        ust |= tmp
        if tmp != set([]):
            unknown.append(i)  # list of implementation reactions with an unknown species
    us = list(ust)  # all species that remain unknown in current (partial) interpretation
    
    # check atomic condition
    atoms = set()
    for k in intrp:
        sps = list(intrp[k].keys())
        if len(sps) == 1 and intrp[k][sps[0]] == 1:
            atoms.add(sps[0])

    atomsleft = list(set(fs) - atoms)
    l = len(atomsleft)
    for assign in permutations(us, l): # works even if l == 0
        # each assign is a tuple of implementation species to be interpreted
        #  as exactly one formal species, matching the order of atomsleft
        # if l == 0 then it.permutations(us, l) == [()], a list with
        #  element which is the zero-length tuple,
        #  so this loop will run exactly once with no change to itmp
        itmp = dict(intrp)
        for i in range(l):
            assert assign[i] not in itmp
            itmp[assign[i]] = Counter({atomsleft[i]: 1})

        sicrntmp = subst(icrn, itmp)
        T = updateT(fcrn, sicrntmp, fs)
        if (not checkT(T)) or any([not T[i][-1] for i in unknown]):
            # either the table is bad, or a reaction we are assuming
            #  is trivial can no longer be trivial
            continue

        ustmp = [sp for sp in us if sp not in assign]
        if not ustmp:
            out = perm(fcrn, icrn, fs, itmp, permcheck, state)
            if out[0]:
                return out
            else:
                state = out[1]
                continue

        n = 0
        a = []
        for i in unknown:
            a.append([])
            for j in ustmp:
                a[n].append(sicrntmp[i][0][j]-sicrntmp[i][1][j])

            a[n].append(0)
            n += 1

        for isp in ustmp:
            itmp[isp] = Counter()

        check = True
        for fsp in fs:
            n = 0
            for j in unknown:
                a[n].pop()
                a[n].append(sicrntmp[j][0][fsp]-sicrntmp[j][1][fsp])
                n += 1

            s = solve(a)
            if s == []:
                check = False
                break
            else:
                for j in range(len(ustmp)):
                    itmp[ustmp[j]][fsp] = s[j]

        if check:
            for isp in ustmp:
                itmp[isp] = itmp[isp] + Counter()
            out = perm(fcrn, icrn, fs, itmp, permcheck, state)
            if out[0]:
                return out
            else:
                state = out[1]
                continue

    return [False, state]

def perm(fcrn, icrn, fs, intrp, permcheck, state):
    """ Check the permissive condition.
    
    Uses the original formal CRN  and original implementation CRN
    (without substitutions).

    Args:
        fcrn: The original formal CRN
        icrn: The original implementation CRN
        fs: The formal species.

    """
    from .crn_bisimulation import subst, updateT, checkT, enumL
    if len(permcheck) == 2:
        permissive_depth = permcheck[1]
        permcheck = permcheck[0]
    else:
        permissive_depth = None

    intr, max_depth, permissive_failure = state

    tr = []
    fr = []
    hasht = set([])

    sicrn = subst(icrn, intrp)
    T = updateT(fcrn, sicrn, fs)
    assert checkT(T) # Assuming this has been checked before calling permissive.
    nulls = [k for k in intrp if not len(list(intrp[k]))] # null species

    def cnstr(s):
        # given formal state s, generate minimal set of impl. states which
        #  interpret to a superset of s
        # = concatenation of any x such that s[0] in m(x) with
        #  all minimal implementations of s - m(x)
        if list(s.keys()) == []:
            yield Counter()
        else:
            s1 = list(s.keys())[0]
            for k in intrp:
                if s1 in intrp[k]:
                    if (s - intrp[k])[s1] >= s[s1]:
                        assert False
                    for out in cnstr(s - intrp[k]):
                        yield Counter({k:1}) + out

    def reactionsearch(s, d):
        # s is the implementation state, d is the depth.
        # try to find (via trivial reactions) a state in which a reaction in fr can fire.
        # fr is a particular formal reaction along with all implementation reactions that interpret to it. 
        # tr is the list of all trivial reactions in the implementation.
        # fail if depth d of trivial reaction steps is exceeded without firing a reaction in fr.
        if permissive_depth and d > permissive_depth:
            return None
        hashee = list(s.elements())
        for i in range(len(hashee)):
            try:
                if hashee[i][0] == 'impl':
                    hashee[i] = hashee[i][1]
            except (IndexError, TypeError):
                pass
        hashee = tuple(sorted(hashee))
        if hashee in hasht:
            return False
        else:
            hasht.add(hashee)
        for i in fr:
            if list((i[0] - s).keys()) == []:
                return True
        ret = False
        for i in tr:
            if list((i[0] - s).keys()) == []:
                t = (s - i[0]) + i[1]
                out = reactionsearch(t, d+1)
                if out:
                    return True
                elif out is None:
                    ret = None
        return ret

    def midsearch(start, goal, pickup, ignore, formal, k):
        # search for a path from start to goal (states of non-null species)
        #  which produces at least pickup null species
        #  assuming it already has infinite copies of everything in ignore
        #  of length at most 2^k
        # if goal is None, the goal is any reaction in fr
        if goal is not None:
            if ignore.issuperset(goal - start):
                return True
            if not interleq(goal, start, intrp):
                return False

        if k == 0:
            if goal is None:
                for rx in fr:
                    if ignore.issuperset(rx[0] - start):
                        return True
                return False
            
            for rx in tr:
                if ignore.issuperset(rx[0] - start):
                    # every element of rx[0] - start (multiset)
                    #  is also an element of ignore (set)
                    # e.g. true on rx[0] - start = {|a,a,b|}, ignore = {a,b,c}
                    if msleq(goal, (start - rx[0]) + rx[1]):
                        return True

        else:
            if midsearch(start,goal,pickup,ignore,formal,k-1):
                return True
            for part in subsets(Counter(pickup)):
                for mid in cnstr(formal):
                    if midsearch(start,mid,set(part),ignore,formal,k-1) \
                       and ((not interleq(start, mid, intrp)) or
                            midsearch(mid,goal,pickup-set(part),ignore,
                                      formal,k-1)):
                        return True

        return False

    def loopsearch(start, formal):
        # search for a path from start to any reaction in fr
        if permissive_depth:
            rounds = math.ceil(math.log(permissive_depth, 2))
        else:
            nequiv = 0
            rounds = 0
            roundequiv = 1
            for point in cnstr(formal):
                nequiv += 1
                if nequiv > roundequiv:
                    rounds += 1
                    roundequiv *= 2

        for parti in enum(len(nulls) + 1,Counter(nulls)):
            part = list(map(set,parti))
            if any([part[i] != set() and part[i+1] == set() \
                    for i in range(len(part) - 2)]):
                continue # avoid redundancy

            pickups = [_f for _f in part[:-1] if _f is not None]
            check1 = True
            place = start
            ignore = set()
            for pickup in pickups:
                check2 = False
                for base in cnstr(formal):
                    if midsearch(place,base,set(),ignore,formal,rounds):
                        if not interleq(place,base,intrp):
                            return True
                        elif midsearch(base,base,pickup,ignore,formal,rounds):
                            check2 = True
                            place = base
                            break

                if not check2:
                    check1 = False
                    break

                ignore |= pickup

            if check1 and midsearch(place,None,set(),ignore,formal,rounds):
                return True

        return False if not permissive_depth else None

    def graphsearch(formal):
        """Check whether every implementation of state "formal" can implement the given formal reaction.

        This function stores for each state the list of states it can reach.
         -- w/o null species (except those producible by loops)
        """
        points = list([[x,set([]),[]] for x in cnstr(formal)])
        # points[i][0] is the ith implementation state
        # points[i][1] is all null species loopable from that state
        # points[i][2][j] is True if points[j][0] is known to be reachable
        #  (from state i without initial null species)
        # exception: points[i][2] is True if a formal reaction is known to
        #  be reachable from state i
        l = len(points) # let's not recalculate this every time...
        rngl = list(range(l)) # same here...
        for i in rngl:
            points[i][2] = l*[False]

        if permissive_depth:
            d = 0
        changed = True
        while changed and ((not permissive_depth) or d < permissive_depth):
            changed = False
            if permissive_depth:
                d = d + 1
            for i in rngl:
                if points[i][2] is not True:
                    for rx in fr:
                        if points[i][1].issuperset(rx[0] - points[i][0]):
                            points[i][2] = True
                            changed = True
                            break

                    if points[i][2] is True:
                        continue

                    for rx in tr:
                        if points[i][1].issuperset(rx[0] - points[i][0]):
                            left = points[i][0] - rx[0]
                            after = left + rx[1]
                            for j in rngl:
                                if msleq(points[j][0],after):
                                    if points[j][2] is True:
                                        points[i][2] = True
                                        changed = True
                                        break

                                    if points[j][2][i]:
                                        s = points[j][1].union(
                                            [x for x in left if x in nulls])
                                        if not s <= points[i][1]:
                                            points[i][1] |= s
                                            changed = True

                                    if not points[i][2][j]:
                                        points[i][2][j] = True
                                        changed = True

                                    for k in rngl:
                                        if (not points[i][2][k]) \
                                           and points[j][2][k]:
                                            points[i][2][k] = True
                                            changed = True

                        if points[i][2] is True:
                            break

        if permissive_depth and changed:
            return None
        return all([p[2] is True for p in points])
    
    n = 0
    # build tr just once
    for i in T:
        if i.pop():
            tr.append(icrn[n])
        n += 1

    for i in range(len(fcrn)):
        # build fr for this formal reaction
        fr = []
        for j in range(len(icrn)):
            if T[j][i]:
                fr.append(icrn[j])

        ist = cnstr(fcrn[i][0])

        if permcheck == "graphsearch":
            out = graphsearch(fcrn[i][0])
            if not out:
                if max_depth == -2:
                    return [False, [intr, max_depth, permissive_failure]]
                if out is False: # proven unreachable
                    max_depth = -1
                else: # arbitrarily specified max search depth reached
                    max_depth = -2
                intr = intrp.copy()
                permissive_failure[0] = fcrn[i]
                permissive_failure[1] = ["Somewhere"]
                return [False, [intr, max_depth, permissive_failure]]

            continue
       
        # At this point, "ist" contains an exhaustive and sufficient list of
        # possible initial implementation states in which the current formal
        # reaction #i must be able to fire (via a series of trivial reactions),
        # Note that we will only want to test states in "ist" that interpret to
        # a state in which #i can fire.

        tested = []  # to avoid testing a superset of some state that's already been tested
        spaceefficient = True # ... but only if we have space to store them
        for j in ist:
            tmp = interpret(j,intrp)
            if msleq(fcrn[i][0], tmp):  # only test if reactants j interpret to allow #i to fire
                t = False
                for k in tested:
                    # will be a 0-length loop if spaceefficient
                    if msleq(k, j):
                        t = True
                        break
                if t:
                    continue
                hasht = set([])
                found = ((permcheck=="loopsearch") and loopsearch(j,fcrn[i][0])) \
                        or ((permcheck=="reactionsearch") and reactionsearch(j,0))
                if not found: # didn't work, but keep track of stuff for user output
                    if max_depth == -2:
                        return [False, [intr, max_depth, permissive_failure]]
                    if found is False: # reaction proven unreachable
                        max_depth = -1
                    else: # arbitrarily specified max search depth reached
                        max_depth = -2
                    intr = intrp.copy()
                    permissive_failure[0] = fcrn[i]
                    permissive_failure[1] = j
                    return [False, [intr, max_depth, permissive_failure]]
                elif not spaceefficient:
                    tested.append(j)
    intr = intrp.copy()
    return [True, intr]

def permissive(fcrn, icrn, fs, intrp, permcheck):
    """ Check the permissive condition.
    
    Uses the original formal CRN  and original implementation CRN
    (without substitutions).

    Args:
        fcrn: The original formal CRN
        icrn: The original implementation CRN
        fs: The formal species.

    """
    from .crn_bisimulation import subst, updateT, checkT, enumL
    permissive_depth = None

    tr = []
    fr = []
    hasht = set([])

    sicrn = subst(icrn, intrp)
    T = updateT(fcrn, sicrn, fs)
    assert checkT(T) # Assuming this has been checked before calling permissive.
    nulls = [k for k in intrp if not len(list(intrp[k]))] # null species

    def cnstr(s):
        # given formal state s, generate minimal set of impl. states which
        #  interpret to a superset of s
        # = concatenation of any x such that s[0] in m(x) with
        #  all minimal implementations of s - m(x)
        if list(s.keys()) == []:
            yield Counter()
        else:
            s1 = list(s.keys())[0]
            for k in intrp:
                if s1 in intrp[k]:
                    if (s - intrp[k])[s1] >= s[s1]:
                        assert False
                    for out in cnstr(s - intrp[k]):
                        yield Counter({k:1}) + out

    def reactionsearch(s, d):
        # s is the implementation state, d is the depth.
        # try to find (via trivial reactions) a state in which a reaction in fr can fire.
        # fr is a particular formal reaction along with all implementation reactions that interpret to it. 
        # tr is the list of all trivial reactions in the implementation.
        # fail if depth d of trivial reaction steps is exceeded without firing a reaction in fr.
        if permissive_depth and d > permissive_depth:
            return None
        hashee = list(s.elements())
        for i in range(len(hashee)):
            try:
                if hashee[i][0] == 'impl':
                    hashee[i] = hashee[i][1]
            except (IndexError, TypeError):
                pass
        hashee = tuple(sorted(hashee))
        if hashee in hasht:
            return False
        else:
            hasht.add(hashee)
        for i in fr:
            if list((i[0] - s).keys()) == []:
                return True
        ret = False
        for i in tr:
            if list((i[0] - s).keys()) == []:
                t = (s - i[0]) + i[1]
                out = reactionsearch(t, d+1)
                if out:
                    return True
                elif out is None:
                    ret = None
        return ret

    def midsearch(start, goal, pickup, ignore, formal, k):
        # search for a path from start to goal (states of non-null species)
        #  which produces at least pickup null species
        #  assuming it already has infinite copies of everything in ignore
        #  of length at most 2^k
        # if goal is None, the goal is any reaction in fr
        if goal is not None:
            if ignore.issuperset(goal - start):
                return True
            if not interleq(goal, start, intrp):
                return False

        if k == 0:
            if goal is None:
                for rx in fr:
                    if ignore.issuperset(rx[0] - start):
                        return True
                return False
            
            for rx in tr:
                if ignore.issuperset(rx[0] - start):
                    # every element of rx[0] - start (multiset)
                    #  is also an element of ignore (set)
                    # e.g. true on rx[0] - start = {|a,a,b|}, ignore = {a,b,c}
                    if msleq(goal, (start - rx[0]) + rx[1]):
                        return True

        else:
            if midsearch(start,goal,pickup,ignore,formal,k-1):
                return True
            for part in subsets(Counter(pickup)):
                for mid in cnstr(formal):
                    if midsearch(start,mid,set(part),ignore,formal,k-1) \
                       and ((not interleq(start, mid, intrp)) or
                            midsearch(mid,goal,pickup-set(part),ignore,
                                      formal,k-1)):
                        return True

        return False

    def loopsearch(start, formal):
        # search for a path from start to any reaction in fr
        if permissive_depth:
            rounds = math.ceil(math.log(permissive_depth, 2))
        else:
            nequiv = 0
            rounds = 0
            roundequiv = 1
            for point in cnstr(formal):
                nequiv += 1
                if nequiv > roundequiv:
                    rounds += 1
                    roundequiv *= 2

        for parti in enum(len(nulls) + 1,Counter(nulls)):
            part = list(map(set,parti))
            if any([part[i] != set() and part[i+1] == set() \
                    for i in range(len(part) - 2)]):
                continue # avoid redundancy

            pickups = [_f for _f in part[:-1] if _f is not None]
            check1 = True
            place = start
            ignore = set()
            for pickup in pickups:
                check2 = False
                for base in cnstr(formal):
                    if midsearch(place,base,set(),ignore,formal,rounds):
                        if not interleq(place,base,intrp):
                            return True
                        elif midsearch(base,base,pickup,ignore,formal,rounds):
                            check2 = True
                            place = base
                            break

                if not check2:
                    check1 = False
                    break

                ignore |= pickup

            if check1 and midsearch(place,None,set(),ignore,formal,rounds):
                return True

        return False if not permissive_depth else None

    def graphsearch(formal):
        """Check whether every implementation of state "formal" can implement the given formal reaction.

        This function stores for each state the list of states it can reach.
         -- w/o null species (except those producible by loops)
        """
        points = list([[x,set([]),[]] for x in cnstr(formal)])
        # points[i][0] is the ith implementation state
        # points[i][1] is all null species loopable from that state
        # points[i][2][j] is True if points[j][0] is known to be reachable
        #  (from state i without initial null species)
        # exception: points[i][2] is True if a formal reaction is known to
        #  be reachable from state i
        l = len(points) # let's not recalculate this every time...
        rngl = list(range(l)) # same here...
        for i in rngl:
            points[i][2] = l*[False]

        if permissive_depth:
            d = 0
        changed = True
        while changed and ((not permissive_depth) or d < permissive_depth):
            changed = False
            if permissive_depth:
                d = d + 1
            for i in rngl:
                if points[i][2] is not True:
                    for rx in fr:
                        if points[i][1].issuperset(rx[0] - points[i][0]):
                            points[i][2] = True
                            changed = True
                            break

                    if points[i][2] is True:
                        continue

                    for rx in tr:
                        if points[i][1].issuperset(rx[0] - points[i][0]):
                            left = points[i][0] - rx[0]
                            after = left + rx[1]
                            for j in rngl:
                                if msleq(points[j][0],after):
                                    if points[j][2] is True:
                                        points[i][2] = True
                                        changed = True
                                        break

                                    if points[j][2][i]:
                                        s = points[j][1].union(
                                            [x for x in left if x in nulls])
                                        if not s <= points[i][1]:
                                            points[i][1] |= s
                                            changed = True

                                    if not points[i][2][j]:
                                        points[i][2][j] = True
                                        changed = True

                                    for k in rngl:
                                        if (not points[i][2][k]) \
                                           and points[j][2][k]:
                                            points[i][2][k] = True
                                            changed = True

                        if points[i][2] is True:
                            break

        if permissive_depth and changed:
            return None
        return all([p[2] is True for p in points])
    
    n = 0
    # build tr just once
    for i in T:
        if i.pop():
            tr.append(icrn[n])
        n += 1

    for i in range(len(fcrn)):
        # build fr for this formal reaction
        fr = []
        for j in range(len(icrn)):
            if T[j][i]:
                fr.append(icrn[j])

        ist = cnstr(fcrn[i][0])

        if permcheck == "graphsearch":
            out = graphsearch(fcrn[i][0])
            if not out:
                if max_depth == -2:
                    return [False, [intr, max_depth, permissive_failure]]
                if out is False: # proven unreachable
                    max_depth = -1
                else: # arbitrarily specified max search depth reached
                    max_depth = -2
                intr = intrp.copy()
                permissive_failure[0] = fcrn[i]
                permissive_failure[1] = ["Somewhere"]
                return [False, [intr, max_depth, permissive_failure]]

            continue
       
        # At this point, "ist" contains an exhaustive and sufficient list of
        # possible initial implementation states in which the current formal
        # reaction #i must be able to fire (via a series of trivial reactions),
        # Note that we will only want to test states in "ist" that interpret to
        # a state in which #i can fire.

        tested = []  # to avoid testing a superset of some state that's already been tested
        spaceefficient = True # ... but only if we have space to store them
        for j in ist:
            tmp = interpret(j,intrp)
            if msleq(fcrn[i][0], tmp):  # only test if reactants j interpret to allow #i to fire
                t = False
                for k in tested:
                    # will be a 0-length loop if spaceefficient
                    if msleq(k, j):
                        t = True
                        break
                if t:
                    continue
                hasht = set([])
                found = ((permcheck=="loopsearch") and loopsearch(j,fcrn[i][0])) \
                        or ((permcheck=="reactionsearch") and reactionsearch(j,0))
                if not found: # didn't work, but keep track of stuff for user output
                    if max_depth == -2:
                        return [False, [intr, max_depth, permissive_failure]]
                    if found is False: # reaction proven unreachable
                        max_depth = -1
                    else: # arbitrarily specified max search depth reached
                        max_depth = -2
                    intr = intrp.copy()
                    permissive_failure[0] = fcrn[i]
                    permissive_failure[1] = j
                    return [False, [intr, max_depth, permissive_failure]]
                elif not spaceefficient:
                    tested.append(j)
    intr = intrp.copy()
    return [True, intr]

def searchr(fcrn, icrn, fs, unknown, intrp, d, permcheck, state, nontriv=False):
    """ Row search: every implementation reaction must interpret to a formal reaction 
        (or be trivial).
    """
    from .crn_bisimulation import subst, updateT, checkT, enumL
    intr, max_depth = state[0:2]
    sicrn = subst(icrn, intrp)
    T = updateT(fcrn, sicrn, fs)
    if not checkT(T):
        # delimiting condition is not satisfied
        yield False
        yield state
        return
    if unknown == []:
        for fsp in fs:
            checkFs = False
            for isp in intrp:
                if intrp[isp] and not intrp[isp] - Counter([fsp]):
                    checkFs = True
                    break
            if not checkFs:
                # atomic condition is not satisfied
                yield False
                yield state
                return
        out = perm(fcrn, icrn, fs, intrp, permcheck, state)
        yield out[0]
        yield out[1]
        return
    if max_depth >= 0 and d > max_depth:
        intr = intrp.copy()
        max_depth = d
    found = False
    min = len(fcrn)+1
    k = -1  # next row reaction we will solve
    nt = 0  # number of possibly trivial reactions according to table
    for i in unknown:
        tmp = T[i].count(True)
        if T[i][-1] == True:
            nt += 1
            tmp -= 1
        if tmp < min and tmp > 0:
            min = tmp
            k = i
    if nt == len(unknown) and not nontriv:  # all unsearched rows could be trivial -- try it!
        out = equations(fcrn, icrn, fs, intrp, permcheck,
                        [intr, max_depth] + state[2:])
        if out[0]:
            yield True
            yield out[1]
            found = True
        else:
            # if we just tried equations and it didn't work, then we
            #  shouldn't try again unless something changes
            nontriv = True
            state = out[1]
            intr, max_depth = state[0:2]
    if k < 0:
        if not found:
            yield False
            yield [intr, max_depth] + state[2:]
        return
    untmp = list(unknown)
    untmp.remove(k)
    if T[k][-1] == True:  # if implementation reaction #k can be trivial, leave it that way and try more rows
        out = searchr(fcrn, icrn, fs, untmp, intrp, d, permcheck,
                      [intr, max_depth] + state[2:], nontriv)
        if next(out):
            if not found:
                found = True
                yield True
            for outintr in out:
                yield outintr
        else:
            state = next(out)
            intr, max_depth = state[0:2]
    n = 0
    for c in range(len(fcrn)): # try to match implementation reaction #k with some formal reaction
        if T[k][c]:        
            ul = sicrn[k][0] - fcrn[c][0]
            kl = list(ul.keys())
            vl = list(ul.values())
            nl = len(kl)
            sl = fcrn[c][0] - sicrn[k][0]
            tmpl = enum(nl, sl, vl)
            ur = sicrn[k][1] - fcrn[c][1]
            kr = list(ur.keys())
            vr = list(ur.values())
            nr = len(kr)
            sr = fcrn[c][1] - sicrn[k][1]
            tmpr = enum(nr, sr, vr)
            for (i,j) in product(tmpl, tmpr):
            # for i in tmpl:
                intrpleft = dict(list(zip(kl, i)))
                # for j in tmpr:
                intrpright = dict(list(zip(kr, j)))
                checkCompatible = True
                for key in intrpleft:
                    if key in intrpright and \
                       any([intrpright[key][fsp] != intrpleft[key][fsp]
                            for fsp in fs]):
                        checkCompatible = False
                        break

                if not checkCompatible:
                    continue

                itmp = intrp.copy()
                itmp.update(intrpleft)
                itmp.update(intrpright)
                out = searchr(fcrn, icrn, fs, untmp, itmp, d+1, permcheck,
                              [intr, max_depth] + state[2:])
                if next(out):
                    if not found:
                        found = True
                        yield True
                    for outintrp in out:
                        yield outintrp
                else:
                    state = next(out)
                    intr, max_depth = state[0:2]
    if not found:
        yield False
        yield [intr, max_depth] + state[2:]
    return




#!/usr/bin/env python
#
#  crnverifier/verifier.py
#  Original source from the Nuskell compiler project
#
import logging
log = logging.getLogger(__name__)

import sys
import signal
import argparse

from crnverifier import __version__
from crnverifier.utils import (parse_crn, 
                               pretty_crn,
                               clean_crn,
                               crnsize,
                               remove_species, 
                               natural_sort)
from crnverifier.crn_bisimulation import (crn_bisimulation_test, 
                                          modular_crn_bisimulation_test)
from crnverifier.pathway_decomposition import (NoFormalBasisError, 
                                               get_formal_basis, 
                                               pathway_decomposition_eq)
from crnverifier.hybrid_notions import (integrated_hybrid_test,
                                        compositional_hybrid_test)

class TimeoutError(Exception):
    pass

def handler(signum, frame):
    raise TimeoutError('Time over')

def limit_runtime(timeout, func, *kargs, **kwargs):
    """ Helper function to exit a function and return None after a specified time.
    """
    signal.signal(signal.SIGALRM, handler)
    signal.alarm(timeout)
    try:
        if timeout:
            log.info(f'Function will abort after {timeout} seconds.')
        ret = func(*kargs, **kwargs)
    except TimeoutError:
        ret = None
    finally:
        tleft = signal.alarm(0)
        if timeout:
            log.info(f'Function needed {int(timeout-tleft)} seconds.')
    return ret

def get_bisimulation_inputs_interactive(constants):
    raise NotImplementedError('no interactive mode support.')

def get_pathway_decomposition_inputs_interactive(constants):
    raise NotImplementedError('no interactive mode support.')

def get_bisimulation_inputs(fCRN, iCRN, interpretation, constants):
    """ Helper function to parse commandline input files.
    """
    fcrn, fs = parse_crn(fCRN, is_file = True)
    fcrn = remove_species(fcrn, constants)
    log.info('Input formal CRN of size {} with formal species: {}\n  {}'.format(
        crnsize(fcrn),
        ", ".join(natural_sort(fs)),
        "\n  ".join(pretty_crn(fcrn))))

    icrn, _ = parse_crn(iCRN, is_file = True)
    icrn = remove_species(icrn, constants)
    log.info('Input implementation CRN:\n  {}'.format("\n  ".join(pretty_crn(icrn))))

    inter = None
    if interpretation:
        inte, _ = parse_crn(interpretation, is_file = True)
        assert all(len(rxn[0]) == 1 for rxn in inte)
        inter = {rxn[0][0]: rxn[1] for rxn in inte}
        log.info('Input interpretation:\n  {}'.format(
                 '\n  '.join(f"{k} -> {' + '.join(v)}" for k, v in natural_sort(inter.items()))))
    return fcrn, icrn, fs, inter

def get_modular_bisimulation_inputs(fCRN, iCRN, interpretation, constants):
    """ Helper function to parse commandline input files.
    """
    fcrns, fs = parse_crn(fCRN, is_file = True, modular = True)
    fcrns = [remove_species(m, constants) for m in fcrns]
    for e, m in enumerate(fcrns, 1):
        log.info('Input formal CRN module: {}\n  {}'.format(e,
            "\n  ".join(pretty_crn(m))))

    icrns, _ = parse_crn(iCRN, is_file = True, modular = True)
    icrn = [remove_species(m, constants) for m in icrns]
    for e, m in enumerate(icrns, 1):
        log.info('Input implementation CRN module: {}\n  {}'.format(e,
            "\n  ".join(pretty_crn(m))))

    inter = None
    if interpretation:
        inte, _ = parse_crn(interpretation, is_file = True)
        assert all(len(rxn[0]) == 1 for rxn in inte)
        inter = {rxn[0][0]: rxn[1] for rxn in inte}
        log.info('Input interpretation:\n  {}'.format(
                 '\n  '.join(f"{k} -> {' + '.join(v)}" for k, v in natural_sort(inter.items()))))
    return fcrns, icrns, fs, inter

def print_bisimulation_outputs(v, i, method, verify_timeout):
    if v is True:
        print(f"Verification result for {method} = {v}.",
              f"The implementation CRN is a correct implementation of the formal CRN.")
    elif v is False:
        print(f"Verification result for {method} = {v}.",
              f"The implementation CRN is not a correct implementation of the formal CRN.")
    elif v is None:
        print(f"No verification result for {method}.",
              f"Verification did not terminate within {verify_timeout} seconds.")

def get_pathway_decomposition_inputs(crn_files, formals, constants):
    """ Helper function to parse commandline input files.
    """
    log.info(f'Formal species: {{{", ".join(natural_sort(formals))}}}')
    log.info(f'Constant species: {{{", ".join(natural_sort(constants))}}}')

    crns = []
    for e, cf in enumerate(crn_files, 1):
        crn, _ = parse_crn(cf, is_file = True)
        crn = remove_species(crn, constants)
        log.info('Input CRN {}:\n  {}'.format(e, "\n  ".join(pretty_crn(crn))))
        crns.append(crn)
    return crns, set(formals)

def add_commandline_arguments(parser):
    correctness = parser.add_argument_group(
            'Arguments for crn-bisimulation & hybrid notions')
    equivalence = parser.add_argument_group(
            'Arguments for pathway decomposition equivalence')

    parser.add_argument("method", 
            action = 'store', 
            choices = ('crn-bisimulation', 'modular-crn-bisimulation', 
                       'pathway-decomposition', 'formal-basis', 
                       'integrated-hybrid', 
                       'compositional-hybrid'),
            help = """Specify verification method.  The method
            *modular-crn-bisimulation* requires a more constrained text input
            format for CRNs: each module must be written on one line. For
            example: "A + B -> C; C -> D + E" written on one line is interpeted
            as one module composed of two reactions. If the formal CRN has n
            modules, and the implementation CRN n+1 modules, then the last
            implementation CRN module is assumed to implement crosstalk, i.e.
            the corresponding formal CRN module is empty. """)

    parser.add_argument("--version", action = 'version', 
                        version = '%(prog)s ' + __version__)
    parser.add_argument("-v", "--verbose", action = 'count', default = 0,
            help = "Print verbose output. -vv increases verbosity level.")
    parser.add_argument("--verify-timeout", type = int, default = 30, metavar = '<int>',
            help = "Specify time in seconds to wait for verification to complete.")
    parser.add_argument("-c", "--constant-species", nargs = '+', default = [], 
            action = 'store', metavar = '<str>', 
            help = """Provide a list of constant species that can be 
            removed before testing equivalence, 
            e.g. fuel species and waste species.""")
    parser.add_argument("--non-modular", action = 'store_true',
            help = """Do not split CRNs into smaller modules when using the
            pathway decomosition method.""")
    parser.add_argument("--profile", action = 'store_true',
            help = """Get some code profiling information (requires statprof-smarkets).""")
 
    correctness.add_argument("-f", "--formal-crn", 
            action = 'store', metavar = '</path/to/file>',
            help = "Read a formal CRN from a file.")
    correctness.add_argument("-i", "--implementation-crn", 
            action = 'store', metavar = '</path/to/file>',
            help = "Read an implementation CRN from a file.")
    correctness.add_argument("-m", "--interpretation", 
            action = 'store', metavar = '</path/to/file>',
            help = """Read an interpretation (a mapping between species in
            typical CRN format) from a file.""")
    correctness.add_argument("--permissive-check",
            action = 'store', metavar = '<str>', default = 'graphsearch',
            choices = ('loopsearch', 'reactionsearch', 'graphsearch'),
            help = "Choose an algorithm to check the permissive condition.")

    equivalence.add_argument("-b", "--crn-files", nargs = '+', default = [],
            action='store', metavar='</path/to/file>',
            help = "Read one or more files that contain a CRN.")
    equivalence.add_argument("-s", "--formal-species", nargs = '+', 
            default = [], action = 'store', metavar = '<str>', 
            help = "List formal species names.")

def main():
    """ Test if an implementation CRN is a correct implementation of a formal CRN.

    An example from Robert's CRN bisimulation code:
    fCRN = "a -> b"
    icrn = "a1 -> b1; x -> a1; x -> b1; y -> b1; y -> a1; x -> a0; a0 -> a1"


    """
    parser = argparse.ArgumentParser(
            formatter_class = argparse.ArgumentDefaultsHelpFormatter)
    add_commandline_arguments(parser)
    args = parser.parse_args()

    logger = logging.getLogger('crnverifier')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    formatter = logging.Formatter('%(levelname)s %(filename)s: %(message)s')
    ch.setFormatter(formatter)
    if args.verbose == 0:
        ch.setLevel(logging.WARNING)
    elif args.verbose == 1:
        ch.setLevel(logging.INFO)
    elif args.verbose == 2:
        ch.setLevel(logging.DEBUG)
    elif args.verbose >= 3:
        ch.setLevel(logging.NOTSET)
    logger.addHandler(ch)

    if args.profile:
        try:
            import statprof
            statprof.start()
        except ImportError as err:
            log.warning('Cannot import statprof module. (pip install statprof-smarkets.)')
            args.profile = False

    if args.method in ('crn-bisimulation', 'integrated-hybrid', 'compositional-hybrid'):
        # TODO: Interactive mode asks for input if fcrn isn't given, etc.
        assert args.formal_crn is not None
        assert args.implementation_crn is not None
        fcrn, icrn, fs, inter = get_bisimulation_inputs(args.formal_crn,
                                                    args.implementation_crn,
                                                    args.interpretation,
                                                    args.constant_species)
        if args.method == 'crn-bisimulation':
            v = limit_runtime(args.verify_timeout,
                              crn_bisimulation_test,
                              fcrn, icrn, fs, 
                              interpretation = inter, 
                              permissive = args.permissive_check)
        elif args.method == 'compositional-hybrid':
            v = limit_runtime(args.verify_timeout,
                              compositional_hybrid_test,
                              fcrn, icrn, fs, inter, 
                              not args.non_modular)
 
        else:
            assert args.method == 'integrated-hybrid'
            v = limit_runtime(args.verify_timeout,
                              integrated_hybrid_test,
                              fcrn, icrn, fs, inter, 
                              not args.non_modular)

        # Unpack the verification results if it wasn't a timeout.
        (v, i) = (None, None) if v is None else v
        if v:
            log.info('Returned interpretation:\n  ' + \
                     '\n  '.join(f"{k} -> {' + '.join(v)}" for k, v in natural_sort(i.items())))
            log.info('Interpreted CRN:\n  {}'.format( 
                     '\n  '.join(pretty_crn(clean_crn(icrn, inter = i)))))
        print_bisimulation_outputs(v, i, args.method, args.verify_timeout)

    elif args.method in ('modular-crn-bisimulation'):
        # TODO: Interactive mode asks for input if fcrn isn't given, etc.
        assert args.formal_crn is not None
        assert args.implementation_crn is not None
        fcrns, icrns, fs, inter = get_modular_bisimulation_inputs(args.formal_crn,
                                                                  args.implementation_crn,
                                                                  args.interpretation,
                                                                  args.constant_species)
        v = limit_runtime(args.verify_timeout,
                          modular_crn_bisimulation_test,
                          fcrns, icrns, fs, 
                          interpretation = inter, 
                          permissive = args.permissive_check)

        # Unpack the verification results if it wasn't a timeout.
        (v, i) = (None, None) if v is None else v
        if v:
            log.info("Returned interpretation:\n  " + \
                     '\n  '.join(f"{k} -> {' + '.join(v)}" for k, v in natural_sort(i.items())))
            for e, icrn in enumerate(icrns, 1):
                log.info('Interpreted CRN module {}:\n  {}'.format(e, 
                         '\n  '.join(pretty_crn(clean_crn(icrn, inter = i)))))
        print_bisimulation_outputs(v, i, args.method, args.verify_timeout)

    elif args.method in ('formal-basis', 'pathway-decomposition'):
        # TODO: Interactive mode asks for input if fcrn isn't given, etc.
        if args.method == 'pathway-decomposition':
            assert len(args.crn_files) > 1
        assert len(args.formal_species)
        crns, fs = get_pathway_decomposition_inputs(args.crn_files,
                                                    args.formal_species,
                                                    args.constant_species)
        if args.method == 'pathway-decomposition':
            v = limit_runtime(args.verify_timeout,
                              pathway_decomposition_eq,
                              crns, fs, not args.non_modular)
            if v is True:
                print(f"Verification result for {args.method} = {v}. The CRNs are equivalent.")
            elif v is False:
                print(f"Verification result for {args.method} = {v}. The CRNs are not equivalent.")
            elif v is None:
                print(f"No verification result for {args.method}.",
                      f"Verification did not terminate within {args.verify_timeout} seconds.")
        elif args.method == 'formal-basis':
            for e, crn in enumerate(crns, 1):
                try:
                    v = limit_runtime(args.verify_timeout,
                                      get_formal_basis,
                                      crn, fs, 
                                      not args.non_modular)

                    if v is None: 
                        print('Timeout, no formal basis found.')
                    else:
                        fbasis, _ = v
                        print('Formal basis {}:\n  {}'.format(e, "\n  ".join(pretty_crn(fbasis))))
                except NoFormalBasisError as err:
                    print("Could not find formal basis {}: {}".format(e, err))
    else:
        raise NotImplementedError(args.method)

    if args.profile:
        statprof.stop()
        statprof.display()

if __name__ == "__main__":
    main()



#
#  crnverifier/hybrid_notions.py
#  Original source from the Nuskell compiler project
#
#  Authors:
#   - Seung Woo Shin (seungwoo.theory@gmail.com)
#   - Stefan Badelt (bad-ants-fleet@posteo.eu)
#
import logging
log = logging.getLogger(__name__)

from itertools import chain

from .utils import assign_crn_species, pretty_crn, natural_sort
from .pathway_decomposition import get_formal_basis, NoFormalBasisError, clean_crn
from .crn_bisimulation import crn_bisimulation_test

class HybridNotionError(Exception):
    pass

def integrated_hybrid_test(*kargs):
    # Wrapper function to use the recommended standard implementation
    return integrated_hybrid_dev1(*kargs)

def compositional_hybrid_test(*kargs):
    # Wrapper function to use the recommended standard implementation
    return compositional_hybrid_dev1(*kargs)

def integrated_hybrid_dev1(fcrn, icrn, fs, inter, modular = True):
    """ Integrated hybrid test - using icrn during path enumeration.
    """
    fcrn = clean_crn(fcrn)
    icrn = clean_crn(icrn)

    # Note: if inter provides interpretations for non-signal species, that's ok here.
    # They will be used as formal species when finding the formal basis.
    fs2 = list(chain(*inter.values()))
    if not all(x in fs2 for x in fs):
        raise HybridNotionError('Some formal species cannot be found in the interpretation.')

    # Interpret waste species as nothing.
    intermediates, wastes, reactive_waste = assign_crn_species(icrn, set(inter.keys()))
    if len(reactive_waste):
        log.debug(f'Reactive waste species detected: {reactive_waste}')
    if len(wastes):
        log.info(f'{len(wastes)} waste species are treated as formal: ' + \
                 f'({", ".join(wastes)})')
        for x in wastes: 
            inter[x] = []

    log.debug('Formal CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(fs)),
        "\n  ".join(pretty_crn(fcrn))))
    log.debug('Implementation CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(inter.keys())),
        "\n  ".join(pretty_crn(icrn))))
    log.debug('Implementation CRN after interpretateion:\n  {}'.format(
        "\n  ".join(pretty_crn(clean_crn(icrn, inter = inter)))))

    try:
        log.debug(f'Formal species to find basis: {set(inter.keys())}')
        fbasis_raw, fbasis_int = get_formal_basis(icrn, 
                                                  set(inter.keys()),
                                                  modular = modular,
                                                  interpretation = inter)
    except NoFormalBasisError as err:
        log.info("Could not find formal basis: {}".format(err))
        return False, inter

    log.debug('Raw formal basis:\n  {}'.format(
        "\n  ".join(pretty_crn(fbasis_raw))))
    log.debug('Interpreted formal basis:\n  {}'.format(
        "\n  ".join(pretty_crn(fbasis_int))))
    return sorted(fcrn) == sorted(clean_crn(fbasis_int)), inter

def integrated_hybrid_dev2(fcrn, icrn, fs, inter, modular = True):
    """ Integrated hybrid test - interpreted icrn during path enumeration.
    """
    fcrn = clean_crn(fcrn)
    icrn = clean_crn(icrn, inter = inter)

    # Note: if inter provides interpretations for non-signal species, that's ok here.
    # They will be used as formal species when finding the formal basis.
    fs2 = list(chain(*inter.values()))
    if not all(x in fs2 for x in fs):
        raise HybridNotionError('Some formal species cannot be found in the interpretation.')

    # Interpret waste species as nothing.
    intermediates, wastes, reactive_waste = assign_crn_species(icrn, set(inter.keys()))
    if len(wastes):
        log.warning(f'{len(wastes)} waste species are treated as formal: ' + \
                    f'({", ".join(wastes)})')
        for x in wastes: 
            inter[x] = []
        # Now update icrn so that wastes are removed!
        icrn = clean_crn(icrn, inter = inter)

    log.debug('Formal CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(fs)),
        "\n  ".join(pretty_crn(fcrn))))
    log.debug('Implementation CRN after interpretateion:\n  {}'.format(
        "\n  ".join(pretty_crn(icrn))))

    try:
        log.debug(f'Formal species to find basis: {set(inter.keys())}')
        fbasis_raw, fbasis_int = get_formal_basis(icrn,
                                                  set(inter.keys()),
                                                  modular = modular)
    except NoFormalBasisError as err:
        log.info("Could not find formal basis: {}".format(err))
        return False, inter

    log.debug('Interpreted formal basis:\n  {}'.format(
        "\n  ".join(pretty_crn(fbasis_raw))))
    return sorted(fcrn) == sorted(fbasis_raw), inter

def compositional_hybrid_dev1(fcrn, icrn, fs, inter, modular = True):
    """ Limited compositional hybrid test - using strong bisimulation.
    """
    fcrn = clean_crn(fcrn)
    icrn = clean_crn(icrn)

    # Note: if inter provides interpretations for non-signal species, that's ok here.
    # They will be used as formal species when finding the formal basis.
    fs2 = list(chain(*inter.values()))
    if not all(x in fs2 for x in fs):
        raise HybridNotionError('Some formal species cannot be found in the interpretation.')

    # Interpret waste species as nothing.
    intermediates, wastes, reactive_waste = assign_crn_species(icrn, set(inter.keys()))
    if len(reactive_waste):
        log.debug(f'Reactive waste species detected: {reactive_waste}')
    if len(wastes):
        log.info(f'{len(wastes)} waste species are treated as formal: ' + \
                 f'({", ".join(wastes)})')
        for x in wastes: 
            inter[x] = []

    log.debug('Formal CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(fs)),
        "\n  ".join(pretty_crn(fcrn))))
    log.debug('Implementation CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(inter.keys())),
        "\n  ".join(pretty_crn(icrn))))

    try:
        log.debug(f'Formal species to find basis: {set(inter.keys())}')
        fbasis_raw, _ = get_formal_basis(icrn, set(inter.keys()), modular = modular)
    except NoFormalBasisError as err:
        log.info("Could not find formal basis: {}".format(err))
        return False, inter

    log.debug('Raw formal basis:\n  {}'.format("\n  ".join(pretty_crn(fbasis_raw))))
    return sorted(fcrn) == sorted(clean_crn(fbasis_raw, inter = inter)), inter

def compositional_hybrid_dev2(fcrn, icrn, fs, inter, modular = True):
    """ Limited compositional hybrid test - using crn_bisimulation_test.
    """
    fcrn = clean_crn(fcrn)
    icrn = clean_crn(icrn)

    # Note: if inter provides interpretations for non-signal species, that's ok here.
    # They will be used as formal species when finding the formal basis.
    fs2 = list(chain(*inter.values()))
    if not all(x in fs2 for x in fs):
        raise HybridNotionError('Some formal species cannot be found in the interpretation.')

    # Interpret waste species as nothing.
    intermediates, wastes, reactive_waste = assign_crn_species(icrn, set(inter.keys()))
    if len(reactive_waste):
        log.debug(f'Reactive waste species detected: {reactive_waste}')
    if len(wastes):
        log.info(f'{len(wastes)} waste species are treated as formal: ' + \
                 f'({", ".join(wastes)})')
        for x in wastes: 
            inter[x] = []

    log.debug('Formal CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(fs)),
        "\n  ".join(pretty_crn(fcrn))))
    log.debug('Implementation CRN with formal species: {}\n  {}'.format(
        ", ".join(natural_sort(inter.keys())),
        "\n  ".join(pretty_crn(icrn))))

    try:
        log.debug(f'Formal species to find basis: {set(inter.keys())}')
        fbasis_raw, _ = get_formal_basis(icrn, set(inter.keys()), modular = modular)
    except NoFormalBasisError as err:
        log.info("Could not find formal basis: {}".format(err))
        return False, inter

    log.debug('Raw formal basis:\n  {}'.format("\n  ".join(pretty_crn(fbasis_raw))))

    v, i = crn_bisimulation_test(fcrn, icrn, fs, 
                                 interpretation = inter, 
                                 permissive = 'whole-graph')
    return v, inter




#
#  crnverifier/crn_bisimulation.py
#  Original source from the Nuskell compiler project
#
#  Authors:
#   - Qing Dong
#   - Robert Johnson
#   - Stefan Badelt
#
import logging
log = logging.getLogger(__name__)

import sys
import copy
import math
import random
from collections import Counter
from itertools import product, combinations, chain

from .utils import pretty_rxn
from .utils import interpret as interpretL

class SpeciesAssignmentError(Exception):
    pass

class EnumSpeciesAssignmentError(Exception):
    pass

class SearchDepthExceeded(Exception):
    pass

# Conversion between internal and external representation of implementation species.
def deformalize(k, fs):
    return f'i{{{k}}}' if k in fs else k

def formalize(intrp):
    intr = {}
    for sp in intrp:
        if sp[:2] == 'i{':
            isp = sp[2:-1]
            intr[isp] = intrp[sp]
        else:
            intr[sp] = intrp[sp]
    return intr

def subst(crn, intrp):
    """ Substitute implementation species with formal species in CRN.  """
    return [[interpretL(part, intrp) for part in rxn] for rxn in crn]

# Utils for list based multiset operations -> it's faster than counters ...
def subsetsL(x, reverse = True):
    """ Generate all (uniqe) multi-subsets of a list x. """
    if reverse:
        return chain(*[combinations(x, l) for l in range(len(x), -1, -1)])
    else:
        return chain(*[combinations(x, l) for l in range(0, len(x) + 1)])

def subtractL(a, b, strict = True): # Multiset a - b.
    """ Returns a new list: a - b. """
    a = a[:]
    for x in b:
        try:
            a.remove(x)
        except ValueError as err:
            if strict:
                raise err
    return a

def is_contained(a, b): # Multiset: a < b
    """ bool: True if list a is fully contained in list b. """
    b = b[:]
    try:
        [b.remove(s) for s in a]
    except ValueError as err:
        return False
    return True

def enumL(n, l, weights = None):
    """ Returns combinations to assign elements of list l to n variables. 

    The weights parameter determines how ofen a corresponding variable n
    appears. I.e if you have n = [x, y], weights = [2, 1], then you must assign
    a twice to the same value in l. In the above example, if l = [A, A, B] then
    the only assignments can be: 
        (x, y) = ((A,), (B,)), or 
        (x, y) = ((), (A, A, B)).

    Raises:
        EnumSpeciecsError('Cannot satisfy the assignment.')

    Yields: 
        tuple: all possible assignents.
    """
    if n == 0:
        yield []
        return

    if weights is None:
        weights = [1 for x in range(n)]
    w = weights[0]

    def ldiv(l, s):
        """ True, if all elements are multiples of s """
        if s == 1 or len(l) == 0:
            return l
        l = Counter(l)
        for k, v in l.items():
            if v % s:
                break
            l[k] = int(v/s)
        else:
            return tuple(l.elements())
        return None
        
    if n == 1:
        l = ldiv(l, w)
        if l is None:
            raise EnumSpeciesAssignmentError
        yield [tuple(l)]
        return

    for ss in set(subsetsL(l)):
        assert is_contained(ss, l)
        wss = ldiv(ss, w)
        if wss is None:
            continue
        try:
            for j in enumL(n-1, subtractL(l, ss), weights[1:]):
                yield [wss] + j
        except EnumSpeciesAssignmentError:
            continue
    return

#TODO(SB): When should we clear the cache?
#SRCACHE = dict()
#def same_reaction_cache(irxn, frxn, fs):
#    global SRCACHE
#    tirxn = tuple(map(tuple, irxn))
#    tfrxn = tuple(map(tuple, frxn))
#    if (tirxn, tfrxn) not in SRCACHE:
#        SRCACHE[(tirxn, tfrxn)] = same_reaction(irxn, frxn, fs)
#    return SRCACHE[(tirxn, tfrxn)]

def same_reaction(irxn, frxn, fs):
    """ Check if irxn *could* be interpreted as frxn.

    This assumes that irxn is already interpreted using the formal species fs.
    SB: Note, that I introduced a new expression expression here, see below.
    """
    ul = subtractL(irxn[0], frxn[0], strict = False) # exess implementation reactants 
    sl = subtractL(frxn[0], irxn[0], strict = False) # exess formal reactants 
    ur = subtractL(irxn[1], frxn[1], strict = False) # exess implementation products
    sr = subtractL(frxn[1], irxn[1], strict = False) # exess formal products

    ir = set(ul) # unassigned implementation reactants 
    fr = set(sl) # exess formal reactants 
    ip = set(ur) # unassingend implementation products
    fp = set(sr) # exess formal products
    if ir & fs or ip & fs:
        # There are excess formal reactants or excess formal products
        # in the implementation reaction, so this reaction cannot
        # implement the formal reaction.
        return False
    if len(fr) and len(ir) == 0:
        # There are excess formal reactants and 
        # no more implementation reactants.
        return False 
    if len(fp) and len(ip) == 0:
        # There are excess formal products and 
        # no more implementation products.
        return False
    if len(fp) and len(ir) and len(ip) and ir == ip:
        # NOTE: This line is new:
        # Example:    A -> B + C
        # cannot be:  A + y -> B + y
        return False

    try: # NOTE: this is new. The very same function we use later to assign species.
        # Left side check
        ul = Counter(ul)
        [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
        next(enumL(len(ul), sl, vl))

        # Right side check
        ur = Counter(ur)
        [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
        next(enumL(len(ur), sr, vr))
    except EnumSpeciesAssignmentError:
        return False
    return True

def trivial_reaction(irxn, fs):
    # Formal reactants and products.
    sl = subtractL(irxn[0], irxn[1], strict = False)
    sr = subtractL(irxn[1], irxn[0], strict = False)
    ir = set(sl) # "non-trivial" reactants
    ip = set(sr) # "non-trivial" products
    if (ir & fs and ip <= fs) or (ip & fs and ir <= fs):
        # If the implementation reactants contain a formal species and
        # all implementation products are (different) formal species, or
        # if the implementation products contain a formal species and
        # all implementation reactants are (different) formal species:
        return False
    else: # Could be trival
        try:
            ul = Counter([x for x in irxn[0] if x not in fs])
            sr = [x for x in sr if x in fs]
            [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
            next(enumL(len(ul), sr, vl))
            ur = Counter([x for x in irxn[1] if x not in fs])
            sl = [x for x in sl if x in fs]
            [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
            next(enumL(len(ur), sl, vr))
        except EnumSpeciesAssignmentError:
            return False
    return True
 
def makeT(fcrn, icrn, fs):
    """ Calculate a table with bool entries.

    Takes an implementation CRN where all implementation species have been
    replaced with the corresponding formal species. Thus, the implementation
    CRN contains a mix of interpreted species and/or uninterpreted species. 

    Returns:
        list[lists[bool]]: A table of bool entries with dimensions: 
            row number = len(icrn), col number = len(fcrn) + 1.
    """
    r = []
    for irxn in icrn:
        rr = [same_reaction(irxn, frxn, fs) for frxn in fcrn]
        rr.append(trivial_reaction(irxn, fs))
        r.append(rr)
    return r

def checkT(T):
    """ Check (partial) interpretation for the delimiting condition.

    Returns:
        bool: True if there is no row or (non-trivial) colum with all False values.
    """
    for row in T:
        if all(not b for b in row):
            return False
    for e, col in enumerate(zip(*T), 1):
        if e != len(T[0]) and all(not b for b in col):
            return False
    return True

def minimal_implementation_states(fstate, inter):
    """ Yields a set of minimal implementation states for a formal state. """
    def gen_supersets():
        if len(fstate) == 0:
            yield []
        else:
            fs = set(fstate).pop()
            fs1 = fstate[0]
            for k in inter:
                if fs1 in inter[k]:
                    for out in minimal_implementation_states(subtractL(fstate, 
                                                                       inter[k], 
                                                                       strict = False), inter):
                        yield [k] + out

    def filter_minimals(lst):
        seen = set()
        for x in lst:
            z = tuple(sorted(x))
            if z in seen:
                continue
            if len(x) == 0:
                yield x
                seen.add(z)
                continue
            for y in combinations(x, len(x)-1):
                if is_contained(fstate, interpretL(y, inter)):
                    break
            else:
                yield x
            seen.add(z)
        return
    return filter_minimals(gen_supersets())

def passes_modularity_condition(bisim, icrn, common_is, common_fs):
    """ Check if a bisimulation satisfies the modularity condition.

    The modularity condition is formulated with respect to subsets of 
    *common implementation* and *common formal* species, where each common
    implementation species must interpet to a common formal species. 
    "Common" should be understood as species wich are *shared* between two or
    more modules. 

    In order to guarantee that the permissive condition is satisfied when
    combining modules, this function asserts whether every implementation
    species can turn into "common" implementation species with the same
    interpretation via trivial reactions (that is the interpretation of the
    reaction must be trival). For details check JDW2019: Definition 3.3
    (Modularity condition).

    The implementation is basically the graphsearch algorithm to check the
    permissive condition, where all "minimal states" are each exactly one
    implementation species.

    Args:
        bisim (dict): An interpretation (= CRN bisimulation).
        icrn (list[lists]): An implementation CRN.
        common_is (set): A set of common implementation species.
        common_fs (set): A set of common formal species.

    Returns:
        bool: True if modularity condition is satisfied.
    """
    # All common implementation species are part of the interpretation.
    assert all(ci in bisim for ci in common_is)
    # All interpretations of common implementation species 
    # are part of common formal species.
    if not all(cf in common_fs for ci in common_is for cf in bisim[ci]):
        log.info('Unknown common formal species.')
        return False

    def Y(s):
        return s in common_is
    def Z(s):
        return len(set(bisim[s]) & common_fs) == 0

    trivial_irxns = [] # trivial reactions
    for rxn in icrn:
        iR = interpretL(rxn[0], bisim) 
        iP = interpretL(rxn[1], bisim) 
        if sorted(iR) == sorted(iP):
            trivial_irxns.append(rxn)

    done = {s for s in bisim if Y(s) or Z(s)}
    todo = {s: [set(), set()] for s in bisim if s not in done} 
    # todo[s] = [reach, produce]

    reset = len(todo) > 0
    while reset:
        reset = False
        for r in list(todo.keys()): # start at any of the non-common intermediates
            [r_reach, r_produce] = todo[r]
            for R, P in trivial_irxns:
                if r not in R:
                    continue
                nR = R[:]
                nR.remove(r)
                if set(nR) > r_produce:
                    # There is at least one more reactant which (as far as we
                    # have seen) cannot be produced by r.
                    continue

                nullsp, mystic = set(), set()
                for p in P:
                    if p in done or (not is_contained(bisim[r], bisim[p])): 
                        # If p is done, then any other p in P is either {} or
                        # it is also a reactant which must be checked
                        # separately. Alternatively, if the interpretation
                        # of products contains the interpretation
                        # of reactants and more, then we can call this one
                        # done and move on to check p.
                        done.add(r)
                        del todo[r]
                        break
                    elif len(bisim[p]) == 0:
                        nullsp.add(p)
                    else:
                        mystic.add(p)
                if r in done:
                    reset = True
                    break
                # Now investigate the mysterious "non-common implementation"
                # products which interpret to a formal species, but are not
                # known to be exit states.
                for p in mystic: 
                    if p not in r_reach:
                        r_reach.add(p)
                        reset = True
                    [p_reach, p_produce] = todo[p]
                    if not (p_reach <= r_reach):
                        # anything reachable by products is reachable by reactants.
                        r_reach |= p_reach
                        reset = True
                    if r in p_reach:
                        # we know r -> p -> r, so all r_nullsp and p_nullsp 
                        # can be produced at infinite amounts.
                        loopable = nullsp | p_produce
                        if not (loopable <= r_produce):
                            r_produce |= loopable
                            reset = True
    return len(todo) == 0

def check_permissive_bruteforce(fcrn, icrn, fs, inter, maxdepth = None):
    """ Check the permissive condition via a brute force path search.
    
    Args:
        fcrn: The formal CRN.
        icrn: The implementation CRN.
        fs: The formal species.
        inter: The interpretation.
        maxdepth: A maximum brute force recursion depth. Defaults to None.

    Raises:
        SearchDepthExceeded: If the hardcoded maxdepth is reached.

    Returns:
        bool: Whether the interpretation passes the permissive condition. 
        info: The search depth if successful, otherwise a tuple (R, S0).
    """
    sicrn = subst(icrn, inter)
    T = makeT(fcrn, sicrn, fs)
    assert checkT(T) # Assuming this has been checked before calling permissive.
    assert all(fs | set(v) == fs for k, v in inter.items())

    if maxdepth is None:
        maxdepth = sys.getrecursionlimit() - 10 # should it be one? should it be two?

    log.debug(f'The implementation CRN:\n' + '\n'.join(
        [f'{e} {t=} {pretty_rxn(sicrn[e])} ({pretty_rxn(icrn[e])})' \
                for e, t in enumerate(T)]))
    log.debug(f'{inter=}')

    def bruteforce(si, seen, trivials, depth = 0):
        if depth > maxdepth:
            raise SearchDepthExceeded('WARNING: brute-force permissive test exceeded recursion depth.')
        st = tuple(sorted(si))
        if st in seen:
            return False
        seen.add(st)
        for irxn in formal_irxns:
            if len(subtractL(irxn[0], si, False)) == 0:
                return True
        for irxn in trivials:
            if len(subtractL(irxn[0], si, False)) == 0:
                nS = subtractL(si, irxn[0]) + irxn[1]
                # Shuffle the crn to reduce the probability of getting stuck in
                # an infinite state space assuming that there is a way out ...
                nexttrivials = trivials[:]
                random.shuffle(nexttrivials)
                if bruteforce(nS, seen, nexttrivials, depth + 1):
                    return True
        return False

    # Find implementation reactions that interpret to trivial reactions.
    trivial_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][-1]]
    for i, frxn in enumerate(fcrn):
        # Find implementation reactions that interpret to this formal reaction.
        formal_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][i]]
        for S0 in minimal_implementation_states(frxn[0], inter):
            assert is_contained(frxn[0], interpretL(S0, inter))
            found = bruteforce(S0, set(), trivial_irxns)
            if not found: 
                return False, (frxn, S0)
    return True, maxdepth

def check_permissive_loopsearch(fcrn, icrn, fs, inter):
    """ The 'loopsearch' algorithm to check the permissive condition. 

    TODO: Acutally this is a very different algorithm then described in the
    paper, so before using this in production, one has to go back and see that
    it is correct and what the time/space complexity of this algorithm is.

    """
    log.warning('This "loopsearch" algorithm is not implemented as described in JDW (2019).')
    sicrn = subst(icrn, inter)
    T = makeT(fcrn, sicrn, fs)
    assert checkT(T) # Assuming this has been checked before calling permissive.
    assert all(fs | set(v) == fs for k, v in inter.items())

    log.debug(f'The implementation CRN:\n' + '\n'.join(
        [f'{e} {t=} {pretty_rxn(sicrn[e])} ({pretty_rxn(icrn[e])})' \
                for e, t in enumerate(T)]))
    log.debug(f'{inter=}')

    # Potential null species.
    nulls = [k for k, v in inter.items() if len(v) == 0]
    # Find implementation reactions that interpret to trivial reactions.
    trivial_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][-1]]

    def reach_with_inf(start, goal = None, pickup = None, ignore = None, k = 0):
        """
        Search for a path from start to goal (states of non-null species) which
        produces at least pickup null species assuming it already has infinite
        copies of everything in ignore of length at most 2^k if goal is None,
        the goal is any reaction in formal_irxns.
        
        Args:
            start: an implementation state corresponding to a formal state.
            goal: a (different) implementation state corresponding to the same
                formal state, or None. If goal is None, then the goal is to find
                any implementation reaction corresponding to the formal reaction.
            pickup: xxx Something related to null species.
            ignore: Assuming there exist infinite copies of those species.
            k: Maximum search depth.
        """
        log.debug(f'CHECK: {start=} {goal=} {ignore=} {pickup=} {k=}')
        if ignore is None:
            ignore = set()
        if pickup is None:
            pickup = []

        #log.debug(f'{start=} {goal=} {ignore=}, {pickup=}, {k=}')
        if goal is None: 
            # Check if a implementation reaction is possible.
            for irxn in formal_irxns:
                # if irxn can happen: 
                if ignore >= set(subtractL(irxn[0], start, False)):
                    log.debug(f'True formal {irxn=}')
                    return True
        elif k == 0:
            # See if start can reach a pickup goal by trivial reactions.
            for irxn in trivial_irxns:
                # if irxn can happen: 
                if ignore >= set(subtractL(irxn[0], start, False)):
                    # and if goal is contained in the results of the trivial rxn.
                    if is_contained(goal+pickup, subtractL(start, irxn[0], False) + irxn[1]):
                        return True
        else:
            if reach_with_inf(start, goal, pickup, ignore, k-1):
                # If we can reach goal with trivial reactions, do it!
                log.debug(f'Shortcut {start=} {goal=} {pickup=} {ignore=} {k=}')
                return True

            # Divide the search into two by choosing a new state mid: 
            #   - find path from start to mid.
            #   - find path from mid to goal.
            for pickpart in map(list, subsetsL(pickup)):
              for mid in M:
                if mid == start or mid == goal: 
                    continue
                log.debug(f'Reach w inf: {start=} {mid=} {goal=} {pickpart=} {k=}')
                if reach_with_inf(start, mid, pickpart, ignore, k-1):
                    if reach_with_inf(mid, goal, subtractL(pickup, pickpart), ignore, k-1):
                        return True
        log.debug(f'FALSE: {start=} {goal=} {ignore=} {pickup=} {k=}')
        return False

    for i, frxn in enumerate(fcrn):
        # Find implementation reactions that interpret to this formal reaction.
        formal_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][i]]
        # From the paper M(r) is the set of minimal implementation states for a formal reaction r.
        M = list(map(sorted, minimal_implementation_states(frxn[0], inter)))

        k, roundequiv = 0, 1
        for nequiv in range(1, len(M)+1):
            if nequiv > roundequiv:
                k += 1
                roundequiv *= 2
        log.debug(f'LS: {k=} to find paths between {M} states: {frxn=}')

        for S0 in M:
            # The formal reaction must be able to happen.
            assert is_contained(frxn[0], interpretL(S0, inter))

            # Is the formal reaction possible w/o null species? 
            if reach_with_inf(S0, None):
                # This state can exit by itself.
                continue

            found = False
            reset = True
            ignore = set()
            # Can we reach a state from where the formal reaction is possible?
            while reset:
                log.debug(f'{reset=}, {found=}, {ignore=}')
                reset = False

                for pickup in map(list, nulls):
                    if set(pickup) & ignore:
                        continue
                    assert len(pickup) != 0
                    log.debug(f'{S0=}, {pickup=}, {ignore=}, {k=} ')

                    if reach_with_inf(S0, S0, pickup, ignore, k):
                        log.debug(f'Pickup! {S0=} {pickup=} {ignore=}.')
                        ignore |= set(pickup)
                        reset = True
                        if reach_with_inf(S0, None, ignore = ignore):
                            found = True
                            break

                for Si in M:
                    if Si == S0: 
                        continue
                    log.debug(f'Testing: {S0=} {Si=} {ignore=} {k=}.')
                    if reach_with_inf(S0, Si, ignore = ignore, k = k):
                        log.debug(f'True: {S0=} {Si=} {ignore=} {k=}.')
                        if reach_with_inf(Si, None, ignore = ignore):
                            log.debug(f'Found exit: {Si=}.')
                            found = True
                            break
                if found is True:
                    break

            if not found:
                return False, S0
    return True, None

def check_permissive_graphsearch(fcrn, icrn, fs, inter):
    """ The 'graphsearch' algorithm to check the permissive condition. 

    Args:
        fcrn: The formal CRN.
        icrn: The implementation CRN.
        fs: The formal species.
        inter: The interpretation.

    Returns:
        bool: Whether the interpretation passes the permissive condition. 
        info: The search depth if successful, otherwise a tuple (R, M).
    """

    sicrn = subst(icrn, inter)
    T = makeT(fcrn, sicrn, fs)
    assert checkT(T) # assuming this has been checked before calling permissive.
    assert all(fs | set(v) == fs for k, v in inter.items())

    log.debug(f'The implementation CRN:\n' + '\n'.join(
        [f'{e} {t=} {pretty_rxn(sicrn[e])} ({pretty_rxn(icrn[e])})' \
                for e, t in enumerate(T)]))
    log.debug(f'{inter=}')

    # Species that interpret to nothing.
    nullsp = set([k for k, v in inter.items() if len(v) == 0])

    # Find implementation reactions that interpret to trivial reactions.
    trivial_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][-1]]

    max_depth = 0
    for i, frxn in enumerate(fcrn):
        # Find implementation reactions that interpret to this formal reaction.
        formal_irxns = [irxn for e, irxn in enumerate(icrn) if T[e][i]]
        # All implementation states which must permit the formal reaction.
        M = list(map(tuple, map(sorted, minimal_implementation_states(frxn[0], inter))))
        done = set() # set of species known to implement the current formal rxn.
        todo = {istate: [set(), set()] for istate in M}
        changed, depth = True, 0
        while changed:
            changed, depth = False, depth + 1
            for istate in M:
                if istate in done:
                    continue
                for irxn in formal_irxns:
                    # If the implementation state can implement the formal reaction.
                    if set(subtractL(irxn[0], istate, False)) <= todo[istate][0]: 
                        done.add(istate)
                        del todo[istate]
                        changed = True
                        break
                if istate in done:
                    continue
                for irxn in trivial_irxns:
                    # If the implementation state allows the implementation reaction.
                    if set(subtractL(irxn[0], istate, False)) <= todo[istate][0]: 
                        # apply the reaction to the current state.
                        nstate = subtractL(list(istate), irxn[0], False) + irxn[1]
                        for istate2 in M:
                            # if one of the other implementation states is reachable
                            if is_contained(istate2, nstate):
                                if istate2 in done:
                                    done.add(istate)
                                    del todo[istate]
                                    changed = True
                                    break
                                if istate in todo[istate2][1]: # I assume its a loop?
                                    s = todo[istate2][0] | (set(nstate) & nullsp)
                                    if not s <= todo[istate][0]:
                                        todo[istate][0] |= s
                                        changed = True
                                if istate2 not in todo[istate][1]:
                                    todo[istate][1].add(istate2)
                                    changed = True
                                if not (todo[istate2][1] <= todo[istate][1]):
                                    todo[istate][1] |= todo[istate2][1]
                                    changed = True
                    if istate in done:
                        break
        if todo: 
            return False, (frxn, M)
        if max_depth < depth: 
            max_depth = depth
    return True, max_depth
        
def passes_permissive_condition(fcrn, icrn, fs, inter, permcheck = 'default'):
    """ Tests an interpretation for the permissive condition.
 
    Args:
        fcrn: The formal CRN.
        icrn: The implementation CRN.
        fs: The formal species.
        inter: The interpretation.
        permcheck: Choice of permissive checker, select from ('default',
                   'grapsearch', 'bruteforce'). Defaults to 'default', which
                   tries the often faster bruteforce algorithm and uses
                   graphsearch whenever that algorithm exceeds the default
                   search depth.

    (TODO): The difficulty of checking the permissive condition scales with the
    number of minimal states (S0 in M) for any given formal reaction r = (R, P),
    which typically scales like (and scales no worse than) n**k where n = |S'|
    (the species of the implementation CRN) and k = |R|. (There may be up to
    order n^k minimal implementation states (M) for some formal reaction r and
    the trajectories by which formal r is implemented may have to pass through
    most or all of S0 in M.  When k is unbounded, checking an interpretation is
    PSPACE-complete.)

    We provide three algorithms: 
        * A bruteforce algorithm with unknown time bounds: a depth-first search
            for a path to implement each formal reaction. Space and time bounds
            are not known, but it appears faster than graphsearch in practice.
        * The graphsearch algorithm which takes poly(n^k) space and time:
            Construct a reachability graph for each formal reaction.
        * A preliminary implementation of the loopsearch algorithm, which
            should run in poly(n,k) space and poly(n^(kn)) time: search for
            "productive loops" using a space-efficient algorithm.
 
    Returns:
        bool: Whether the interpretation passes the permissive condition. 
        info: The search depth if successful, otherwise a tuple (r, M).
    """
    log.debug(f'Checking permissive condition using {permcheck=}.')
    if permcheck == 'graphsearch':
        passes, info = check_permissive_graphsearch(fcrn, icrn, fs, inter)
    elif permcheck == 'loopsearch':
        passes, info = check_permissive_loopsearch(fcrn, icrn, fs, inter)
    elif permcheck == 'bruteforce':
        passes, info = check_permissive_bruteforce(fcrn, icrn, fs, inter)
    elif permcheck == 'default':
        try:
            passes, info = check_permissive_bruteforce(fcrn, icrn, fs, inter)
        except SearchDepthExceeded as err:
            log.debug('Switching to graphsearch permissive checker.')
            passes, info = check_permissive_graphsearch(fcrn, icrn, fs, inter)
    else:
        raise SystemExit(f'Unknown algorithm for permissive condition: {permcheck}')
    return passes, info

def passes_delimiting_condition(fcrn, icrn, fs, inter):
    """ Tests an interpretation for the delimiting condition.

    Note that this function is actually never used, because typically you want
    to keep the table T. In any case, it returns the correct result.
    """
    sicrn = subst(icrn, inter)
    T = makeT(fcrn, sicrn, fs)
    return checkT(T)

def passes_atomic_condition(inter, fs):
    """ Tests an interpretation for the atomic condition.

    For every formal species there exists an implementation species which
    interprets to it.  
    """
    return all(any([f] == v for v in inter.values()) for f in fs)

def in_order(inter, order, unassigned):
    """ Check if an interpretation follows an order of assignments.

    Args:
        inter (dict): An interpretation dictionary.
        order (list): A list of implementation species. (They have to be
            assigned in the interpretation.)
        unassigend (set): A set of implementation species who's order has
            not yet been assigend.

    Returns:
        bool, info: True if interpretation is in order, updated order.
    """
    allthere = True
    for isp in order:
        if allthere:
            allthere = (isp in inter)
        elif isp in inter:
            return False, None
    if len(unassigned):
        for isp in inter:
            if isp in unassigned:
                order.append(isp)
                unassigned.remove(isp)
    return True, (order, unassigned)

def search_row(fcrn, icrn, fs, intrp, order = None, seen = None, depth = 0, permcheck = 'default'):
    """ Find full interpretations matching every irxn to one frxn or trxn.

    This "row search" finds all valid combinations of 
        implementation reaction -> formal reaction or 
        implementation reaction -> trivial reaction.
    Typically this function is called after search_column, which means we
    already have a partial interpretation that satisfies the delimiting
    condition. However, the delimiting condition may still break when an
    additional implementation reaction is introduced, so it is always necessary
    to check all three conditions.

    Yields:
        Iterator over all correct interpretations (CRN bisimulations).
    """
    log.debug(f'Searching row at {depth=}'.center(80-(2*depth), '~'))
    log.debug(f'Partial interpretation {intrp=}')
    log.debug(f'Ordered species {order=}')

    if order is None:
        itot = set().union(*[set().union(*rxn[:2]) for rxn in icrn])
        iordered = list(itot & set(intrp))  # must be an ordered list.
        iunassigned = itot - set(intrp)     # better a set.
    else:
        iordered, iunassigned = order
    valid, order = in_order(intrp, iordered, iunassigned)
    if not valid:
        raise SpeciesAssignmentError('Unordered input.')
    
    # We are in correct order, but there may still be duplicates...
    sicrn = subst(icrn, intrp)
    T = makeT(fcrn, sicrn, fs)
    if not checkT(T):
        log.debug(f'Delimiting condition not satisfied.')
        raise SpeciesAssignmentError('Delimiting condition not satisfied.')

    # The row indices where unassigned species can be found, but we 
    # remove duplicates of implementation reactions that solve for the
    # same set of implementation species!
    unknown = {tuple(sorted((set(sirxn[0]) | set(sirxn[1])) - fs)): i \
                for i, sirxn in enumerate(sicrn)} 
    unknown = [v for k, v in unknown.items() if len(k)]

    if unknown == []:
        if not passes_atomic_condition(intrp, fs):
            log.debug(f'Atomic condition not satisfied.')
            raise SpeciesAssignmentError('Atomic condition not satisfied.')
        correct, info = passes_permissive_condition(fcrn, icrn, fs, intrp, permcheck)
        if correct:
            log.debug(f'Permissive condition satisfied ({info=}).')
            yield intrp
            return
        log.debug(f'Permissive condition not satisfied ({info=}).')
        raise SpeciesAssignmentError('Permissive condition not satisfied.')

    unknown = sorted(unknown, key = lambda x: (T[x][-1], T[x][:-1].count(True)))
    # Start with the row with the minimal number of True's (excluding trivial reactions)
    log.debug(f'{unknown=} \n' + '\n'.join(
        [f'{e} {t=} {pretty_rxn(sicrn[e])} ({pretty_rxn(icrn[e])})' \
                for e, t in enumerate(T)]))

    later = dict() # less promising partial interpretations ...
    def search(later, n = 0, m = None):
        # use this to define a search hierarchy when sending results into the
        # next round. Currently, n and m are start and stop in a range of 
        # how many species interpet to a single species.
        while later:
            if m and n > m:
                break
            if n in later:
                for (inext, depth) in later[n]:
                    try:
                        for ir in search_row(fcrn, icrn, fs, inext, order,
                                                 depth = depth + 1, 
                                                 permcheck = permcheck):
                            yield ir
                    except SpeciesAssignmentError:
                        continue
                del later[n]
            n += 1

    alltriv = all(r is False for k in unknown for r in T[k][:-1]) 
    for k in unknown:
        irxn = sicrn[k] 
        if T[k][-1] is True: # Assign a trivial reaction.
            log.debug(f'Interpret: {pretty_rxn(irxn)} => trivial {fs=}')
            # Unassigned reactants and products.
            ul = Counter([x for x in irxn[0] if x not in fs])
            ur = Counter([x for x in irxn[1] if x not in fs])
            # If this reaction has no unassigned species, then we shouldn't be here!
            assert len(ul) or len(ur)
            # Formal reactants and products.
            sl = [x for x in subtractL(irxn[0], irxn[1], False) if x in fs]
            sr = [x for x in subtractL(irxn[1], irxn[0], False) if x in fs]
            log.debug(f'Interpret: {pretty_rxn(irxn)} => {sl=}, {sr=} | {ul=} {ur=}')
            [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
            [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
            tmpl = enumL(len(ul), sr, vl)
            tmpr = enumL(len(ur), sl, vr)
            for i, j in product(tmpl, tmpr):
                intrpleft = {k: tuple(sorted(v)) for k, v in zip(kl, i)}
                intrpright = {k: tuple(sorted(v)) for k, v in zip(kr, j)}
                for key in set(intrpleft) & set(intrpright):
                    if any([intrpleft[key][fsp] != \
                            intrpright[key][fsp] for fsp in fs]):
                        # Incompatible dictionaries!
                        break
                else:
                    inext = intrp.copy()
                    imove = tuple(sorted(chain(intrpleft.items(), intrpright.items())))
                    log.debug(f'Interpret: {imove}')
                    inext.update({k: list(v) for k, v in imove})
                    level = max(len(v) for k, v in imove)
                    later[level] = (later.get(level, [])) + [(inext, depth)]

            if len(ul) and len(ur):
                # An implementation reaction where species on both sides are
                # not interpreted yet, may also be interpreted as a trivial
                # reaction including an additional (single!!) formal species.
                for atom in fs:
                    log.debug(f'Interpret: {pretty_rxn(irxn)} => matching {atom=}')
                    [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
                    tmpl = enumL(len(ul), [atom]+sr, vl)
                    [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
                    tmpr = enumL(len(ur), [atom]+sl, vr)
                    try:
                        for i, j in product(tmpl, tmpr):
                            intrpleft = {k: tuple(sorted(v)) for k, v in zip(kl, i)}
                            intrpright = {k: tuple(sorted(v)) for k, v in zip(kr, j)}
                            for key in set(intrpleft) & set(intrpright):
                                if any([intrpleft[key][fsp] != \
                                        intrpright[key][fsp] for fsp in fs]):
                                    # Incompatible dictionaries!
                                    break
                            else:
                                inext = intrp.copy()
                                imove = tuple(sorted(chain(intrpleft.items(), intrpright.items())))
                                log.debug(f'Interpret: {imove}')
                                inext.update({k: list(v) for k, v in imove})
                                level = max(len(v) for k, v in imove)
                                later[level] = (later.get(level, [])) + [(inext, depth)]
                    except EnumSpeciesAssignmentError:
                        log.debug(f'Reaction {pretty_rxn(irxn)} => matching {atom=} ' + \
                                  f'was not compatible with trivial {fs=}!')

        for c, frxn in enumerate(fcrn): # Assign a formal reaction.
            if not T[k][c]:
                continue
            assert not alltriv
            log.debug(f'Interpret: {pretty_rxn(irxn)} => {pretty_rxn(frxn)}')
            # left 
            ul = Counter(subtractL(irxn[0], frxn[0], False))
            sl = subtractL(frxn[0], irxn[0], False)
            # right
            ur = Counter(subtractL(irxn[1], frxn[1], False))
            sr = subtractL(frxn[1], irxn[1], False)
            if len(ul) or len(ur):
                [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
                [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
                tmpl = enumL(len(ul), sl, vl)
                tmpr = enumL(len(ur), sr, vr)
                for i, j in product(tmpl, tmpr):
                    intrpleft = {k: tuple(sorted(v)) for k, v in zip(kl, i)}
                    intrpright = {k: tuple(sorted(v)) for k, v in zip(kr, j)}
                    for key in set(intrpleft) & set(intrpright):
                        if any([intrpleft[key][fsp] != \
                                intrpright[key][fsp] for fsp in fs]):
                            # Incompatible dictionaries!
                            break
                    else:
                        inext = intrp.copy()
                        inext.update({k: list(v) for k, v in intrpleft.items()})
                        inext.update({k: list(v) for k, v in intrpright.items()})
                        level = max(len(v) for k, v in chain(intrpleft.items(), 
                                                             intrpright.items()))
                        later[level] = (later.get(level, [])) + [(inext, depth)]
            for i in search(later, n = 0, m = 1):
                yield i
        for i in search(later, n = 0, m = 2):
            yield i
    for i in search(later, n = 2):
        yield i 
    assert len(later) == 0
    return

def search_column(fcrn, icrn, fs = None, intrp = None, unknown = None, depth = 0):
    """ Find all partial interpretations matching every frxn to one irxn.

    This "column search" finds all combinations where all formal reactions have
    one matching implementation reaction. For example, if there is one formal
    reaction and three compatible implementation reactions, then this function
    returns three interpretations that have exactly one reaction interpreted.
    Note that every result of the column search must pass the delimiting
    condition, but not necessarily atomic or permissive condition.

    Yields:
        Iterator over partial interpretations satisfying the delimiting condition.
    """
    if fs is None:
        fs = set().union(*[set().union(*rxn[:2]) for rxn in fcrn])
    if intrp is None:
        intrp = dict()
    log.debug(f'Searching column at {depth=}'.center(40-(2*depth), '*'))
    log.debug(f'Partial interpretation {intrp=}')

    sicrn = subst(icrn, intrp)
    T = makeT(fcrn, sicrn, fs)
    if not checkT(T):
        #log.debug(f'Delimiting condition not satisfied with {intrp=}')
        raise SpeciesAssignmentError('Delimiting condition not satisfied.')

    if unknown is None:
        unknown = [i for i in range(len(fcrn))]

    # Start with the unknown column with the minimal number of True's.
    # I.e. the implementation rxn that can be assigned to the least formal reactions.
    unknown = sorted(unknown, key = lambda x: sum(T[j][x] for j in range(len(sicrn))))

    if len(unknown) == 0: 
        # Note, we cannot test for the atomic condition at this point, because
        # some trivial reactions may be needed to satisfy the atomic condition.
        yield intrp
        return

    later = dict() # less promising partial interpretations ...
    def search(later, n = 0, m = None):
        # use this to define a search hierarchy when sending results into the
        # next round. Currently, n and m are start and stop in a range of 
        # how many species interpet to a single species.
        while later:
            if m and n > m:
                break
            if n in later:
                for (inext, unext, depth) in later[n]:
                    try:
                        for isuccess in search_column(fcrn, icrn, fs, 
                                                      inext, unext, depth + 1):
                            yield isuccess
                    except SpeciesAssignmentError:
                        continue
                del later[n]
            n += 1

    for c in unknown:
        frxn = fcrn[c]
        unext = [u for u in unknown if u != c]
        for k, irxn in enumerate(sicrn):
            if not T[k][c]:
                continue
            log.debug(f'Interpret: {pretty_rxn(irxn)} => {pretty_rxn(frxn)}')
            # left 
            ul = Counter(subtractL(irxn[0], frxn[0], False))
            sl = subtractL(frxn[0], irxn[0], False)
            # right
            ur = Counter(subtractL(irxn[1], frxn[1], False))
            sr = subtractL(frxn[1], irxn[1], False)

            if len(ul) or len(ur):
                [kl, vl] = zip(*ul.items()) if len(ul) else [[], []]
                [kr, vr] = zip(*ur.items()) if len(ur) else [[], []]
                tmpl = enumL(len(ul), sl, vl)
                tmpr = enumL(len(ur), sr, vr)
                for i, j in product(tmpl, tmpr):
                    intrpleft = {k: tuple(sorted(v)) for k, v in zip(kl, i)}
                    intrpright = {k: tuple(sorted(v)) for k, v in zip(kr, j)}
                    for key in set(intrpleft) & set(intrpright):
                        if any([sorted(intrpleft[key][fsp]) != \
                                sorted(intrpright[key][fsp]) for fsp in fs]):
                            # Incompatible dictionaries!
                            break
                    else:
                        inext = intrp.copy()
                        inext.update({k: list(v) for k, v in intrpleft.items()})
                        inext.update({k: list(v) for k, v in intrpright.items()})
                        level = max(len(v) for k, v in chain(intrpleft.items(), 
                                                             intrpright.items()))
                        later[level] = (later.get(level, [])) + [(inext, unext, depth)]
            else: # Just in case the full interpretation is provided.
                later[0] = (later.get(0, [])) + [(intrp.copy(), unext, depth)]
            for isuccess in search(later, n = 0, m = 1):
                yield isuccess
        for isuccess in search(later, n = 1, m = 2):
            yield isuccess
    for isuccess in search(later, n = 2):
        yield isuccess
    assert len(later) == 0
    return

def crn_bisimulations(fcrn, icrn, 
                      interpretation = None,
                      formals = None, 
                      permissive = 'default'):
    """ Iterate over all crn bisimulations.

    for e, bisim in enumerate(crn_bisimulations(fcrn, icrn), 1):
        print(f'Bisimulation {e} = {bisim}')

    Note that it is possible to provide more formal species than actually
    appear in the formal CRN. This is because you can still return trivial
    reactions involving this formal species. Hence, the algorithm does try to
    match trivial reactions between formal species, but only those that involve
    one (additional) formal species at a time.

    Limitations: The CRN bisimulation algorithm finds *all* CRN bisimulations
        compatible with a fully enforced partial interpretation. That means if
        you want to constrain the CRN bisimulations to some subset where e.g
        m(x) = A or m(x) = B, but m(x) != C, then you must leave m(x)
        unconstrained and filter the results afterwards.

    Args:
        fcrn (list): A formal CRN.
        icrn (list): An implementation CRN.
        interpretation (dict, optional): A (partial) interpretation with the
            format interpretation[is] = list[fs,..]. Defaults to None. 
        formals (set, optional): The set of formal species. 
            Defaults to None: all species in the formal CRN.
        permissive (string, optional): Choice of permissive checker, select
            from ('default', 'grapsearch', 'bruteforce'). Defaults to
            'default', which tries the often faster bruteforce algorithm and
            uses graphsearch whenever that algorithm exceeds the default search
            depth.

    Yields:
        Iterator over all correct interpretations (CRN bisimulations).
    """
    if interpretation is None:
        interpretation = dict()
    if formals is None:
        formals = set().union(*[set().union(*rxn[:2]) for rxn in fcrn])

    # Now this is a funny one. An interpretation contains a "formal" species
    # that does not appear in the formal CRN and is also not present in
    # extra "formals". May happen if you are not careful combining modules!
    assert all(set(v) & formals == set(v) for v in interpretation.values())

    log.debug(f'Finding CRN bisimulations using {permissive} algorithm:')
    log.debug('Original formal CRN:')
    [log.debug('  {}'.format(pretty_rxn(r))) for r in fcrn]
    log.debug('Original implementation CRN:')
    [log.debug('  {}'.format(pretty_rxn(r))) for r in icrn]
    log.debug(f'Formal species: {formals}')

    if icrn == [] and fcrn != []:
        return

    if permissive not in ['default', 'graphsearch', 'loopsearch', 'bruteforce']:
        raise ValueError('Unknown option: {}'.format(
            'the permissive test should be {}'.format(
            '"graphsearch", "loopsearch", or "bruteforce".')))
    new = []
    for [r, p] in icrn:
        nr = [deformalize(k, formals) for k in r]
        np = [deformalize(k, formals) for k in p]
        assert sorted(nr) != sorted(np)
        new.append([nr, np])
    icrn = new
    log.debug('Internal implementation CRN:')
    [log.debug('  {}'.format(pretty_rxn(r))) for r in icrn]
    inter = {deformalize(k, formals): v for k, v in interpretation.items()
                                                } if interpretation else {}
    log.debug('Internal interpretation:')
    [log.debug('  {}: {}'.format(k,v)) for (k, v) in inter.items()]

    log.info(f'Searching for bisimulation.')
    try: # Needed if you start with a wrong partial interpretation.
        seen = set()
        for parti in search_column(fcrn, icrn, formals, inter):
            hpart = tuple(sorted(tuple([k,tuple(sorted(v))]) for k, v in parti.items()))
            if hpart in seen:
                continue
            seen.add(hpart)
            try:
                for bisim in search_row(fcrn, icrn, formals, parti, 
                                        permcheck = permissive):
                    yield formalize(bisim)
            except SpeciesAssignmentError:
                continue
        log.debug(f'Done after checking {len(seen)} partial interpretations.')
    except SpeciesAssignmentError:
        log.info(f'Unable to satisfy delimiting condition with initial partial interpretation.')
    return

def modular_crn_bisimulation_test(fcrns, icrns, 
                                  formals = None, 
                                  interpretation = None, 
                                  permissive = 'default'):
    """ Check if a modulular CRN bisimulation exists. 

    Note: There are a few modifications to the original source:
        - the arguments to check CRN bisimulation for each module changed:
            - only the formal species present in the module are passed on
        - the modularity condition input changed
            - isc (former ispCommon) are now all implementation species that
              are both in the current module and in at least one other module.
            - fsc are all formal species that are both in the current module
              and in at least one other module.
        - it is still possible to provide one extra implementation module. In
          that case, all implementation reactions have to interpret to trivial
          reactions of the form {} -> {}, UNLESS, there are interpreted
          implementation species in the last module. Those may be trivial
          reactions involving only that species.
    
    Raises:
        NotImplementedError: If there are shared implementation species between
        modules for which no partial interpretation is provided, then there is 
        no guarantee that two modules can be combined. 

    Args:
        see crn_bisimulations()

    Returns:
        [bool, dict/None]: Whether a modular CRN bisimulation exists, the bisimulation or None.
    """
    # Let's make a copy here to avoid modification of the partial interpretation.
    inter = {k : v for k, v in interpretation.items()} if interpretation else {}

    # Identify common implementation species.
    ispc = dict() # Store for every implementation species in which module it appears: 
    for e, module in enumerate(icrns, 1):
        mspecies = set().union(*[set().union(*rxn[:2]) for rxn in module])
        for isp in mspecies:
            ispc[isp] = ispc.get(isp, []) + [e]
    log.debug(f'ispc = {ispc}')

    # Identify common formal species.
    fspc = dict() # Store for every formal species in which module it appears: 
    for e, module in enumerate(fcrns, 1):
        mspecies = set().union(*[set().union(*rxn[:2]) for rxn in module])
        for fsp in mspecies:
            fspc[fsp] = fspc.get(fsp, []) + [e]
    log.debug(f'fspc = {fspc}')

    if formals is None:
        formals = list(fspc.keys())

    if len(fcrns) == len(icrns)-1:
        # If one extra icrn is provided, that means that all implementation
        # rxns have to interpret trivial rxns of the form "{} -> {}". 
        # UNLESS, we have dedicated implementation species in the last module
        # which are assigned to formal species in the partial interpretation!
        # Those may be trivial reactions involving only that species.
        fcrns.append([])
        imspecies = set().union(*[set().union(*rxn[:2]) for rxn in icrns[-1]])
        fmspecies = set().union(*[set().union(inter.get(isp, [])) for isp in imspecies])
        for fsp in fmspecies:
            fspc[fsp] = fspc.get(fsp, []) + [len(fcrns)]

    # Now a quick check if there are common implementation species without interpretation:
    for e, (fcrn, icrn) in enumerate(zip(fcrns, icrns), 1):
        mfs = {k for k in formals if e in fspc[k]}
        minter = {k: v for k, v in inter.items() if e in ispc[k]}
        fsc = {f for f, m in fspc.items() if e in m and len(m) > 1} # S0
        isc = {i for i, m in ispc.items() if e in m and len(m) > 1} # S'0
        if not all(i in minter for i in isc):
            raise NotImplementedError('Modular CRN bisimulation: ' + \
                f'please provide an interpretation for all shared implementation species: {isc}')

    for e, (fcrn, icrn) in enumerate(zip(fcrns, icrns), 1):
        # Prepare inputs for crn bisimulation of this module
        mfs = {k for k in formals if e in fspc[k]}
        minter = {k: v for k, v in inter.items() if e in ispc[k]}
        for bisim in crn_bisimulations(fcrn, icrn, 
                                       interpretation = minter, 
                                       formals = mfs, 
                                       permissive = permissive):
            # Get all formal and implementation species that are in
            # common with at least one other module.
            fsc = {f for f, m in fspc.items() if e in m and len(m) > 1} # S0
            isc = {i for i, m in ispc.items() if e in m and len(m) > 1} # S'0
            if passes_modularity_condition(bisim, icrn, isc, fsc):
                inter.update(bisim)
                break
            log.debug(f'Skipping non-modular bisimulation: {bisim}')
        else:
            return False, None
    return True, inter

def crn_bisimulation_test(fcrn, icrn, 
                          formals = None, 
                          interpretation = None,
                          permissive = 'default'):
    """ Backward compatible CRN bisimulation interface.

    Args:
        see crn_bisimulations()

    Returns:
        [bool, dict/None]: Whether a modular CRN bisimulation exists, the bisimulation or None.
    """
    iterator = crn_bisimulations(fcrn, icrn, 
                                 interpretation = interpretation,
                                 formals = formals, 
                                 permissive = permissive)

    try:
        bisim = next(iterator)
        return True, bisim
    except StopIteration:
        return False, None

